from typing import Tuple

from prometheus_client import CollectorRegistry
from prometheus_client import push_to_gateway as _push_to_gateway

from .model import *

PROM_URL = "http://promethous.logis.com"
PROM_GATEWAY_URL = "http://prom-gateway.logis.com"
registry = CollectorRegistry()


def push_to_prometheus(job: str, **kwargs):
    return _push_to_gateway(PROM_GATEWAY_URL, job=job, registry=registry, **kwargs)


from influxdb_client.client.exceptions import InfluxDBError


class BatchingCallback(object):

    def success(self, conf: Tuple[str, str, str], data: str):
        print(f"Written batch: {conf}, data: {data}")

    def error(self, conf: Tuple[str, str, str], data: str, exception: InfluxDBError):
        print(f"Cannot write batch: {conf}, data: {data} due: {exception}")

    def retry(self, conf: Tuple[str, str, str], data: str, exception: InfluxDBError):
        print(
            f"Retryable error occurs for batch: {conf}, data: {data} retry: {exception}"
        )


callback = BatchingCallback()



import logging
from typing import Generic, Literal, Optional, Union

import requests
from pydantic import BaseModel

from .base import DEFAULT_PYDANTIC_MODEL_CONFIG, T


class ApiError(BaseModel):
    code: int = -1
    message: str = "错误"


class ApiResponse(BaseModel, Generic[T]):

    success: bool
    data: Optional[T] = None
    error: Union[ApiError, Optional[str]] = None
    message: Optional[str] = None
    extra: Optional[dict] = None

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    @staticmethod
    def positive(data: T, **kwargs):
        return ApiResponse[T](success=True, data=data, **kwargs)

    @staticmethod
    def negative(error: Union[ApiError, str], **kwargs):
        return ApiResponse[None](success=False, error=error, **kwargs)

    @staticmethod
    def from_http_response(
        r: requests.Response, content_type: Optional[Literal["json", "text"]] = None
    ):
        ct = content_type or r.headers.get("content-type") or ""
        try:
            data = r.json() if "json" in ct else r.text
        except Exception as e:
            logging.warning("解析响应体失败: %s", e)
            data = r.text
        return (
            ApiResponse.positive(data)
            if r.status_code == 200
            else ApiResponse.negative(
                ApiError(code=r.status_code, message=str(r.status_code))
            )
        )



from datetime import datetime
from typing import Dict, Optional

from pydantic import BaseModel


class PlatformInfo(BaseModel):
    signature: Optional[str] = None
    url: str


class VersionInfo(BaseModel):
    version: str  # 版本号，必选字符串
    notes: str  # 更新说明，必选字符串
    pub_date: datetime  # 发布时间，自动解析ISO 8601格式（如2026-02-02T10:00:00Z）
    platforms: Dict[str, PlatformInfo]


class UpdateResponse(BaseModel):
    current_version: str
    version_info: VersionInfo
    need_update: bool



from abc import ABC, abstractmethod
from typing import Generic, List, TypeVar


class INode(ABC):
    pass


class IEdge(ABC):

    @abstractmethod
    def get_weight(self, *args, **kwargs) -> float:
        """
        获取边的权重
        """
        pass

    @abstractmethod
    def is_directed(self) -> bool:
        """
        是否有向边
        """
        pass


E = TypeVar("E", bound=IEdge)
V = TypeVar("V")


class IPathGraph(ABC, Generic[E, V]):
    """
    路径关系图
    """

    @abstractmethod
    def edges(self, *args, **kwargs) -> List[E]:
        """
        所有的边
        """
        pass

    @abstractmethod
    def nodes(self, *args, **kwargs) -> List[V]:
        """
        所有的节点
        """
        pass



import asyncio
from pathlib import Path
from typing import Any, Dict, Generic, TypeVar, Union

import simpy
from networkx import DiGraph
from pydantic import AliasChoices
from pyee import EventEmitter

from logis.data_type import DEFAULT_PYDANTIC_MODEL_CONFIG, ComponentId, Point, TmpId
from logis.iface import Storable
from logis.util.websocket_util import (
    ClientConnection,
    ServerConnection,
    is_websocket_closed,
)

from .agent import *
from .component import *

T = TypeVar("T")

LocationType = Union[Point, Storable]


class SimMetadata(BaseModel):
    """
    仿真元数据
    """

    sim_id: Optional[str] = Field(
        validation_alias=AliasChoices("sim_id", "simId"),
        default=None,
    )
    model_path: Optional[Path] = Field(
        validation_alias=AliasChoices("model_path", "modelPath"), default=None
    )
    # 历史逻辑仅支持单个仿真的同步模式，新逻辑新增完全异步模式
    mode: Optional[Literal["sync", "async"]] = None

    user_id: Optional[str] = Field(
        validation_alias=AliasChoices("user_id", "user"), default=None
    )

    client_version: Optional[str] = Field(
        validation_alias=AliasChoices("client_version", "version"), default=None
    )

    server_version: Optional[str] = Field(
        validation_alias=AliasChoices("server_version", "serverVersion"), default=None
    )

    send_step_info_realtime: Optional[bool] = Field(
        None,
        validation_alias=AliasChoices(
            "send_step_info_realtime", "sendStepInfoRealtime"
        ),
    )

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG


class SimConfig(SimMetadata):
    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    json_path: Optional[str] = None
    excel_path: Optional[str] = None
    result_path: Optional[str] = None

    # 实际最长运行时间，单位s，默认不限时
    max_run_time: Optional[float] = None
    # 最大模拟时间，默认不限时
    max_sim_time: Optional[float] = None

    def resolve_model_name(self) -> Optional[str]:
        from logis.biz.sim.path_util import resolve_model_name

        return resolve_model_name(self.model_path)

    def resolve_result_path(self):
        if self.result_path:
            return Path(self.result_path)
        from logis.biz.sim.path_util import resolve_result_path

        return resolve_result_path(self.model_path)

    def resolve_excel_path(self) -> Optional[Path]:
        if self.excel_path:
            return Path(self.excel_path)
        from logis.biz.sim.path_util import resolve_excel_path

        return resolve_excel_path(self.model_path)

    def resolve_json_path(self) -> Optional[Path]:
        if self.json_path:
            return Path(self.json_path)
        from logis.biz.sim.path_util import resolve_json_path

        return resolve_json_path(self.model_path)


class OldSimEvent(SimMetadata):
    """
    旧版仿真事件结构体，包含了仿真过程中产生的事件可能用到的所有字段
    """

    # 下面的都是历史结构体，所有的字段融合在一起，后续逐步使用泛型代替
    entity_id: Optional[TmpId] = Field(
        validation_alias=AliasChoices("entity_id", "agent_id", "id"), default=None
    )
    # 智能体类型
    entity_type: Optional[ComponentId] = None
    # 起始位置
    start_pos: Optional[List[float]] = None
    start_point: Optional[Point] = None
    # 结束位置
    end_pos: Optional[List[float]] = None
    end_point: Optional[Point] = None
    # 时长
    duration: Optional[Union[int, float]] = None
    # FIXME:名字叫货架ID，但实际上存的全是stock_id
    payload_id: Optional[TmpId] = Field(
        validation_alias=AliasChoices("shelf_id", "payload_id"), default=None
    )
    rack_id: Optional[TmpId] = None
    # 货物载体ID
    carrier_id: Optional[TmpId] = None

    init_stock: Optional[bool] = None
    stock: Optional[Any] = None
    stock_id: Optional[TmpId] = None

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG


class SimEvent(OldSimEvent, Generic[T]):
    """
    仿真过程中产生的事件

    此结构体起初是个大杂烩，包含了仿真过程中产生的事件可能用到的所有字段，后来逐步改造成泛型

    TODO: 这里只所以还继承OldSimEvent，是因为历史代码里有很多地方已经用了SimEvent\\
    实际上是OldSimEvent，后续会逐步替换掉并去除这里的继承
    """

    source_id: Optional[TmpId] = Field(
        validation_alias=AliasChoices("source_id", "src"), default=None
    )

    # 仿真事件发生的时间
    time: Optional[Union[int, float]] = Field(default=None)
    # 事件类型
    event_type: Optional[str] = Field(
        validation_alias=AliasChoices("eventType", "event_type", "type"), default=None
    )
    # 事件数据
    data: Optional[T] = None
    extra: Optional[Dict[str, Any]] = None

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG
    # 真实时间戳
    timestamp: Optional[float] = None

    def to_config(self):
        return SimConfig(
            sim_id=self.sim_id,
            model_path=self.model_path,
            mode=self.mode,
            send_step_info_realtime=self.send_step_info_realtime,
        )


class SimContext(SimMetadata):
    """
    仿真上下文
    """

    # 正常完成的信号
    __finish_event__: Optional[simpy.Event] = None
    # 终止仿真的信号
    __terminate_event__: Optional[simpy.Event] = None
    env: simpy.Environment = Field(default_factory=simpy.Environment)
    event_emitter: EventEmitter = Field(default_factory=EventEmitter)
    sim_config: Optional[SimConfig] = None
    errors: List[str] = Field(default_factory=list)
    network: DiGraph = Field(default_factory=DiGraph)
    ws_conn: Union[ClientConnection, ServerConnection, None] = None
    event_loop: Optional[asyncio.BaseEventLoop] = None

    extra: Optional[Dict[str, Any]] = None

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    @property
    def terminate_event(self) -> simpy.Event:
        """
        仿真终止事件
        """
        assert self.env, "环境未初始化"
        env = self.env
        if self.__terminate_event__ is None:
            self.__terminate_event__ = env.event()
        return self.__terminate_event__

    @property
    def finish_event(self) -> simpy.Event:
        """
        仿真完成事件
        """
        assert self.env, "环境未初始化"
        env = self.env
        if self.__finish_event__ is None:
            self.__finish_event__ = env.event()
        return self.__finish_event__

    def trigger_finish_event_if_not(self):
        """
        如果仿真完成事件还没有被触发，就触发它
        """
        if not self.finish_event.triggered:
            self.finish_event.succeed()

    def trigger_terminate_event_if_not(self):
        """
        如果仿真终止事件还没有被触发，就触发它
        """
        if not self.terminate_event.triggered:
            self.terminate_event.succeed()

    def is_terminated(self):
        """
        判断仿真是否已被终止
        """
        return self.terminate_event.triggered

    def is_finished(self):
        """
        判断仿真是否已完成
        """
        return self.finish_event.triggered

    def is_over(self):
        """
        判断仿真是否结束
        """
        if self.ws_conn and (is_websocket_closed(self.ws_conn)):
            return True
        return self.is_finished() or self.is_terminated()

    def resolve_sim_id(self) -> Optional[str]:
        """
        优先读取sim_config中的id，其次是本结构体中的id
        """
        return (self.sim_config.sim_id if self.sim_config else None) or self.sim_id

    def resolve_model_path(self) -> Optional[Path]:
        """
        优先读取sim_config中的model_path，其次是本结构体中的model_path
        """
        return (
            self.sim_config.model_path if self.sim_config else None
        ) or self.model_path

    def resolve_model_name(self) -> Optional[str]:
        path = self.resolve_model_path()
        return path.name if path else None



from pathlib import Path
from typing import Generic, Optional, TypeVar, Union

from pydantic import BaseModel

from .base import DEFAULT_PYDANTIC_MODEL_CONFIG

V = TypeVar("V")


class ConfigItem(BaseModel, Generic[V]):
    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    namespace: str = ""

    label: Union[str, None] = None
    key: str
    value: Optional[V] = None
    value_type: Optional[str] = None
    disabled: bool = False


class AppConfig(BaseModel):
    """
    应用常用配置
    """

    app_name: str
    root_dir: Optional[Path] = None
    runtime_root_dir: Optional[Path] = None
    data_dir: Optional[Path] = None
    logs_dir: Optional[Path] = None
    sidecar_dir: Optional[Path] = None
    conf_dir: Optional[Path] = None

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG



from logis.util.lambda_util import generate_all, invoke


def test_generator_result():

    def c():
        return None

    def a():
        yield 1
        yield 2
        yield 3
        return 4

    def b():
        yield from a()
        yield 5
        yield 6
        return 8

    result = generate_all(b())
    assert result.return_value == 8
    assert result.yield_values == [1, 2, 3, 5, 6]

    result = generate_all(a())
    assert result.return_value == 4
    assert result.yield_values == [1, 2, 3]

    result = invoke(c)
    assert result.return_value is None
    assert result.yield_values == []


def test_invoke():
    def a():
        return 1

    async def b():
        return 2

    result = invoke(a)
    assert result.return_value == 1
    assert result.yield_values == []

    result = invoke(b)
    assert result.return_value == 2
    assert result.yield_values == []



from abc import ABC, abstractmethod
from typing import List, Optional, Tuple, Union

from pydantic import BaseModel

from logis.data_type import DEFAULT_PYDANTIC_MODEL_CONFIG

from .order import IOrder


class OrderPickingConfig(BaseModel):
    """
    订单拣选配置项
    """

    # 拣选工人能同时处理的订单数量
    picker_parallel_orders: Optional[int] = None

    # # 分波次的方式
    # wave_method: Optional[Literal["订单数量", "时间段"]] = None
    # # 所选合并订单方式的值
    # value_of_wave_method: Optional[Union[UTime, int]] = None

    # 按照订单优先级排序
    sort_by_priority: bool = False

    # 智能体专用
    merge_by_sku_when_fetching: bool = True

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG


class OrderPickingStrategy(ABC):
    """
    订单拣选策略
    1. 合并size均为1时，退化为按单拣选
    2. 合并size均>1时，即是合并拣选
    3. 合并size>=1时，即是混合拣选
    """

    @abstractmethod
    def on_merge(self, *orders: Tuple[IOrder], **kwargs):
        """
        当订单发生合并时的钩子方法
        """
        pass

    @abstractmethod
    def merge(
        self, orders: List[IOrder], config: Union[OrderPickingConfig, None] = None
    ) -> List[IOrder]:
        """
        订单合并逻辑
        Args:
            orders: 待合并的订单列表
            config: 订单拣选配置
        Returns:
            合并后的订单列表
        """
        if not orders:
            return []
        if config and config.sort_by_priority:
            orders = sorted(orders, key=lambda od: od.get_priority())
        return orders



from datetime import datetime
from typing import Optional, Set

from pydantic import BaseModel, Field

from logis.data_type import DEFAULT_PYDANTIC_MODEL_CONFIG
from logis.task import TaskId


class WaveInfo(BaseModel):
    """
    波次信息
    """

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG
    wave_start_time: Optional[datetime] = None
    wave_id: Optional[str] = None
    order_ids: Set[str] = Field(default_factory=set)
    task_ids: Set[TaskId] = Field(default_factory=set)



import pandas as pd
import tabulate


def to_pretty_string(df: pd.DataFrame) -> str:
    """
    将DataFrame转换为易读的字符串格式
    Args:
        df: 输入的DataFrame
    Returns:
        易读的字符串格式
    """
    return tabulate.tabulate(df, headers="keys", tablefmt="fancy_grid")



import math
from typing import List, Optional, Tuple, Union

from pyecharts import options as opts


class GridLayout:
    """
    智能网格布局计算器
    根据图表数量和布局要求，自动计算每个图表的GridOpts位置

    示例：
        calculator = GridLayoutCalculator(total_items=6, rows=2, cols=3)
        grid_opts_list = calculator.calculate()

        或自动计算行列：
        calculator = GridLayoutCalculator(total_items=5)
        grid_opts_list = calculator.calculate()  # 自动计算为2行3列
    """

    def __init__(
        self,
        total_items: int,
        rows: Optional[int] = None,
        cols: Optional[int] = None,
        width: str = "1200px",
        height: str = "800px",
        margin_top: str = "5%",
        margin_bottom: str = "5%",
        margin_left: str = "5%",
        margin_right: str = "5%",
        spacing_x: str = "2%",
        spacing_y: str = "2%",
    ):
        """
        初始化布局计算器

        Args:
            total_items: 图表总数量
            rows: 指定行数（如为None则自动计算）
            cols: 指定列数（如为None则自动计算）
            width: 总画布宽度
            height: 总画布高度
            margin_*: 边距（百分比或像素）
            spacing_*: 图表间距（百分比）
        """
        self.total_items = total_items
        self.rows = rows
        self.cols = cols
        self.width = width
        self.height = height
        self.margin_top = margin_top
        self.margin_bottom = margin_bottom
        self.margin_left = margin_left
        self.margin_right = margin_right
        self.spacing_x = spacing_x
        self.spacing_y = spacing_y

        # 自动计算行列数（如果未指定）
        if self.rows is None and self.cols is None:
            self.rows, self.cols = self._auto_calculate_grid()
        elif self.rows is None:
            self.rows = math.ceil(self.total_items / self.cols)
        elif self.cols is None:
            self.cols = math.ceil(self.total_items / self.rows)

    def _auto_calculate_grid(self) -> Tuple[int, int]:
        """自动计算最佳的行列数"""
        # 如果是1个图表，使用1x1
        if self.total_items <= 1:
            return 1, 1

        # 尝试找到一个接近正方形的布局
        sqrt = math.sqrt(self.total_items)
        rows = math.floor(sqrt)
        cols = math.ceil(self.total_items / rows)

        # 如果布局太窄（列数远大于行数），调整一下
        if cols / rows > 2:
            rows = math.ceil(sqrt)
            cols = math.ceil(self.total_items / rows)

        return rows, cols

    def _parse_percentage(self, value: str) -> float:
        """解析百分比字符串为小数"""
        if isinstance(value, str) and value.endswith("%"):
            return float(value.rstrip("%")) / 100
        elif isinstance(value, (int, float)):
            return float(value) / 100
        return 0.0

    def _to_percentage(self, value: float) -> str:
        """将小数转换为百分比字符串"""
        return f"{value * 100:.1f}%"

    def get_all_center_radius(self):
        """
        根据行列自动计算中心点、半径的百分值
        """
        r, c = self.rows, self.cols

        radius = 100 / max(r, c)
        y_per = 100 / (r * 2)
        y_centers = [(i + 1) * y_per for i in range(0, r * 2, 2)]
        x_per = 100 / (c * 2)
        x_centers = [(i + 1) * x_per for i in range(0, c * 2, 2)]

        count = 0
        for y_center in y_centers:
            for x_center in x_centers:
                yield dict(center=(x_center, y_center), radius=radius)
                count += 1
                if count >= self.total_items:
                    return

    def get_all_grid_opts(self) -> List[opts.GridOpts]:
        """
        计算所有图表的GridOpts配置

        Returns:
            按照顺序排列的GridOpts列表
        """
        # 解析边距和间距
        margin_top = self._parse_percentage(self.margin_top)
        margin_bottom = self._parse_percentage(self.margin_bottom)
        margin_left = self._parse_percentage(self.margin_left)
        margin_right = self._parse_percentage(self.margin_right)
        spacing_x = self._parse_percentage(self.spacing_x)
        spacing_y = self._parse_percentage(self.spacing_y)

        # 计算可用空间
        available_width = 1.0 - margin_left - margin_right
        available_height = 1.0 - margin_top - margin_bottom

        # 计算每个单元格的宽度和高度
        cell_width = (available_width - (self.cols - 1) * spacing_x) / self.cols
        cell_height = (available_height - (self.rows - 1) * spacing_y) / self.rows

        grid_opts_list = []

        for i in range(self.total_items):
            # 计算行列索引
            row = i // self.cols
            col = i % self.cols

            # 计算位置
            pos_left = margin_left + col * (cell_width + spacing_x)
            pos_top = margin_top + row * (cell_height + spacing_y)

            # 创建GridOpts
            grid_opts = opts.GridOpts(
                pos_left=self._to_percentage(pos_left),
                pos_top=self._to_percentage(pos_top),
                pos_right=self._to_percentage(1.0 - (pos_left + cell_width)),
                pos_bottom=self._to_percentage(1.0 - (pos_top + cell_height)),
                width=self._to_percentage(cell_width),
                height=self._to_percentage(cell_height),
            )

            grid_opts_list.append(grid_opts)

        return grid_opts_list

    def get_layout_info(self) -> dict:
        """获取布局信息"""
        return {
            "total_items": self.total_items,
            "rows": self.rows,
            "cols": self.cols,
            "actual_items": min(self.total_items, self.rows * self.cols),
            "empty_cells": max(0, self.rows * self.cols - self.total_items),
        }



from datetime import datetime
from typing import Any, Generic, Optional, TypeVar

from pydantic import BaseModel
from pyee import EventEmitter
from pyee.base import Handler

from logis.ctx import Context
from logis.data_type import DEFAULT_PYDANTIC_MODEL_CONFIG

T = TypeVar("T")


class EventContext(BaseModel, Generic[T]):
    # 事件所携带的数据
    payload: Optional[T] = None
    # 事件源ID
    source_id: Optional[str] = None
    # 事件目标ID
    target_id: Optional[Any] = None
    # 事件名称
    name: str
    # 附加数据
    extra: Any = None

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG


class EventBus(Context):
    """
    事件总线，支持监听、发出事件
    """

    emitter = EventEmitter()

    @classmethod
    def get_event_emitter(cls) -> EventEmitter:
        return cls.emitter

    @classmethod
    def emit(cls, event_name: str, *args, **kwargs):
        cls.get_event_emitter().emit(event_name, *args, **kwargs)

    @classmethod
    def on(cls, event_name: str, func: Handler):
        cls.get_event_emitter().on(event_name, func)

    @classmethod
    def add_listener(cls, event_name: str, listener: Handler):
        cls.get_event_emitter().add_listener(event_name, listener)


class EventBaseModel(BaseModel, Generic[T]):
    """
    事件基础结构
    """

    # 设计为x.y.z的形式
    event_type: str

    value: Optional[T] = None

    source_id: Optional[str] = None

    # 事件发生的时间
    event_time: float

    created_at: Optional[datetime] = None



from pathlib import Path

from setuptools_scm import _config, get_version

ROOT_DIR = Path(__file__).parent.parent

version_file = ROOT_DIR / "logis/auto_generated_version.py"

get_version(
    root=ROOT_DIR,
    # relative_to=__file__,
    version_scheme=_config.DEFAULT_VERSION_SCHEME,
    write_to=version_file,
)



from abc import ABC, abstractmethod

from fastapi import APIRouter, FastAPI


class RouteRegister(ABC):

    @abstractmethod
    def get_router(self, **kwargs) -> APIRouter:
        pass

    def register(self, app: FastAPI, **kwargs):
        app.include_router(self.get_router(**kwargs))



from typing import Generic, List, Optional, TypeVar

from pydantic import AliasChoices, BaseModel, Field

from .base import DEFAULT_PYDANTIC_MODEL_CONFIG

T = TypeVar("T")


class Pager(BaseModel):
    page_base: Optional[int] = Field(
        default=0, validation_alias=AliasChoices("page_base", "pageBase")
    )
    page_size: int = Field(
        default=100, validation_alias=AliasChoices("page_size", "pageSize")
    )
    page_at: int = Field(
        default=0, validation_alias=AliasChoices("page_at", "pageAt", "page")
    )
    total: Optional[int] = None

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    @property
    def offset(self) -> int:
        return max(0, self.page_at - self.page_base) * self.page_size

    @property
    def limit(self) -> int:
        return self.page_size


class PageResult(Pager, Generic[T]):
    items: List[T]

    @staticmethod
    def of_page(page: Pager, items: List[T], total: int) -> "PageResult[T]":
        return PageResult[T](
            items=items,
            total=total,
            page_at=page.page_at,
            page_base=page.page_base,
            page_size=page.page_size,
        )

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG


class TableQuery(Pager):
    database: Optional[str] = None
    table: str

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG



import logging
import logging.config
from logging import StreamHandler
from logging.handlers import RotatingFileHandler

from logis._log import LoggerBuilder, get_default_dict_config_builder
from logis._log.util import DictConfigBuilder


def do_log(log: logging.Logger):
    log.debug("This is a debug message")
    log.info("This is an info message")
    log.warning("This is a warning message")
    log.error("This is an error message")
    log.critical("This is a critical message")
    try:
        1 / 0
    except Exception as e:
        log.exception("This is an exception message: %s", e)


LOGGER_NAME = "test_logger"

demo_logger = (
    LoggerBuilder()
    .dir("./logs")
    .name(LOGGER_NAME)
    .level(logging.INFO)
    .add_handler(
        StreamHandler,
        filters=[
            lambda record: record.levelno >= logging.WARNING,
        ],
    )
    .add_handler(RotatingFileHandler, filename="test")
    .build()
)


def test_logger_builder():
    do_log(demo_logger)


def test_dict_config_builder():
    dict_config = DictConfigBuilder().logger(demo_logger).build()
    print(dict_config)
    logging.config.dictConfig(dict_config)
    log = logging.getLogger(LOGGER_NAME)
    do_log(log)


def test_default_dict_config_builder():
    dict_config = get_default_dict_config_builder().build()
    print(dict_config)
    logging.config.dictConfig(dict_config)
    log = logging.getLogger("app")
    do_log(log)


if __name__ == "__main__":
    test_logger_builder()



import logging
import logging.handlers
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from uuid import uuid4

from colorlog import ColoredFormatter

from logis.util.pkg_util import get_class_full_path

from .fmt import LoggerFormat
from .model import HandlerDictConfig, LoggerDictConfig


def format_level(level: Union[int, str]) -> int:
    """
    将日志级别统一调整为logging内置常量
    """
    if isinstance(level, str):
        level = level.upper()
        level = getattr(logging, level)
    return level


def format_filename(
    filename: str,
    log_dir: Optional[Path] = None,
    ensure_suffix: Optional[str] = ".log",
) -> str:
    """
    格式化日志文件名，确保文件名以指定后缀结尾。
    如果未指定后缀，则默认使用 ".log"。
    如果未指定日志根目录，则直接返回文件名。
    如果指定了日志根目录，则确保目录存在，并返回绝对路径。
    """
    if ensure_suffix and not filename.endswith(ensure_suffix):
        filename += ensure_suffix
    if log_dir is None:
        return filename
    log_dir = Path(log_dir)
    log_dir.mkdir(parents=True, exist_ok=True)
    return str(log_dir / filename)


def format_filename_if_necessary(
    handler: logging.Handler,
    log_dir: Optional[Path] = None,
    ensure_suffix: Optional[str] = ".log",
) -> None:
    """
    如果给定的handler是文件处理器类型，则格式化其文件名。
    """
    if not handler:
        return
    if isinstance(handler, logging.FileHandler):
        handler.baseFilename = format_filename(
            handler.baseFilename, log_dir, ensure_suffix
        )


def add_formatter_if_not(handler: logging.Handler) -> None:
    """
    如果给定的handler上尚未配置formatter， 则为其设置默认的formatter。
    """
    if not handler:
        return
    if handler.formatter is not None:
        return
    if isinstance(handler, logging.FileHandler):
        handler.setFormatter(logging.Formatter(LoggerFormat.DEFAULT.value))
    else:
        handler.setFormatter(ColoredFormatter(LoggerFormat.COLORED_DEFAULT.value))


def get_dict_config_tmpl():
    """
    获取默认的 logging.config.dictConfig 字典配置模板
    """
    return {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": defaultdict(dict),
        "filters": defaultdict(dict),
        "handlers": defaultdict(dict),
        "loggers": defaultdict(dict),
    }


def formatter_to_dict(formatter: logging.Formatter) -> Dict[str, Any]:
    """
    将 logging.Formatter 实例转换为 logging.config.dictConfig 兼容的字典配置
    """
    return {
        "class": get_class_full_path(formatter.__class__),
        "format": formatter._fmt,
        "datefmt": formatter.datefmt,
    }


def get_name_attr(obj: Any) -> str:
    """
    获取对象的名称， 如果对象没有名称，则返回其类型名。
    """
    return getattr(obj, "name", None)


def handler_to_dict(handler: logging.Handler) -> Dict[str, Any]:
    """
    将 logging.Handler 实例转换为 logging.config.dictConfig 兼容的字典配置
    """
    formatter = getattr(handler, "formatter", None)
    formatter_name = getattr(formatter, "name", None)
    handler_args: Dict[str, Any] = {
        "class": get_class_full_path(handler.__class__),
        "level": handler.level,
        "formatter": formatter_name,
        "filters": [
            get_name_attr(f)
            for f in handler.filters
            if isinstance(f, logging.Filter) and get_name_attr(f) is not None
        ],
    }
    if isinstance(handler, logging.FileHandler):
        handler_args["filename"] = handler.baseFilename
        handler_args["mode"] = handler.mode
        handler_args["encoding"] = handler.encoding
        handler_args["delay"] = handler.delay
    elif isinstance(handler, logging.StreamHandler):
        try:
            fileno = handler.stream.fileno()
            handler_args["stream"] = (
                "ext://sys.stdout" if fileno == 1 else "ext://sys.stderr"
            )
        except (AttributeError, OSError):
            handler_args["stream"] = "ext://sys.stdout"
    else:
        raise ValueError(f"Unsupported handler type: {handler}")

    if isinstance(handler, logging.handlers.RotatingFileHandler):
        handler_args["maxBytes"] = handler.maxBytes
        handler_args["backupCount"] = handler.backupCount
    elif isinstance(handler, logging.handlers.TimedRotatingFileHandler):
        handler_args["when"] = handler.when
        handler_args["interval"] = handler.interval
        handler_args["backupCount"] = handler.backupCount

    return handler_args


class DictConfigBuilder:
    """
    支持从对象实例中提取配置，
    并将其转换为 logging.config.dictConfig 兼容的字典配置
    1. 内置LoggerFormat枚举的所有格式
    """

    def dir(self, log_dir: str):
        self._log_dir = Path(log_dir)
        return self

    def __init__(self):
        self._log_dir: Optional[Path] = None
        self.__dict_config__ = get_dict_config_tmpl()
        self.__loggers__: Dict[str, Dict[str, Any]] = self.__dict_config__["loggers"]
        self.__handlers__: Dict[str, Dict[str, Any]] = self.__dict_config__["handlers"]
        self.__filters__: Dict[str, Dict[str, Any]] = self.__dict_config__["filters"]
        self.__formatters__: Dict[str, Dict[str, Any]] = self.__dict_config__[
            "formatters"
        ]

        for e in LoggerFormat:
            if "color" in e.name.lower():
                formatter = ColoredFormatter(fmt=e.value)
            else:
                formatter = logging.Formatter(fmt=e.value)
            self.formatter(formatter, name=e.formatter_name())

    def build(self):
        return self.__dict_config__

    def filter(self, f: logging.Filter, name: Optional[str] = None):
        """
        将 logging.Filter 实例转换为 logging.config.dictConfig 兼容的字典配置

        返回格式符合 Python logging 官方 dict 配置规范，支持直接用于 dictConfig()
        """
        if not isinstance(f, logging.Filter):
            logging.warning("only logging.Filter instance supported now: %s", f)
            return self
        name = name if name is not None else get_name_attr(f)
        setattr(f, "name", name)
        filter_dict: dict = self.__filters__[name]
        filter_dict.update({"()": get_class_full_path(f.__class__)})
        return self

    def handler(self, handler: logging.Handler, name: Optional[str] = None):
        """
        将 logging.Handler 实例转换为 logging.config.dictConfig 兼容的字典配置

        返回格式符合 Python logging 官方 dict 配置规范，支持直接用于 dictConfig()
        """
        name = name if name is not None else handler.name
        if name is None:
            name = str(uuid4())
        handler.name = name

        format_filename_if_necessary(
            handler, log_dir=self._log_dir, ensure_suffix=".log"
        )

        for f in handler.filters:
            if get_name_attr(f) is None:
                setattr(f, "name", str(uuid4()))

        if formatter := getattr(handler, "formatter", None):
            formatter_name = getattr(formatter, "name", None)
            if formatter_name is None:
                formatter_name = str(uuid4())
            setattr(formatter, "name", formatter_name)

        handler_dict = handler_to_dict(handler)
        formatter = getattr(handler, "formatter", None)
        if formatter is not None:
            self.formatter(formatter)
        for f in handler.filters:
            self.filter(f)
        self.__handlers__[name].update(handler_dict)
        return self

    def formatter(self, formatter: logging.Formatter, name: Optional[str] = None):
        """
        将 logging.Formatter 实例转换为 logging.config.dictConfig 兼容的字典配置

        返回格式符合 Python logging 官方 dict 配置规范，支持直接用于 dictConfig()
        """
        if formatter is None:
            return self
        name = name or getattr(formatter, "name", None)
        setattr(formatter, "name", name)
        formatter_dict = formatter_to_dict(formatter)
        self.__formatters__[name].update(formatter_dict)
        return self

    def logger(self, logger: logging.Logger):
        """
        将 logging.Logger 实例转换为 logging.config.dictConfig 兼容的字典配置

        返回格式符合 Python logging 官方 dict 配置规范，支持直接用于 dictConfig()
        """
        if logger is None:
            return self
        logger_dict: dict = self.__loggers__[logger.name]
        handler_ids = []
        for handler in logger.handlers:
            self.handler(handler)
            handler_ids.append(handler.name)

        logger_dict.update(
            {
                "level": logger.getEffectiveLevel(),
                "propagate": "yes" if logger.propagate else "no",
                "handlers": handler_ids,
            }
        )

        return self

    def logger_dict(self, config: LoggerDictConfig):
        self.__loggers__[config.name].update(config.model_dump(exclude={"name"}))
        return self

    def handler_dict(self, config: HandlerDictConfig):
        config.filename = format_filename(
            config.filename, log_dir=self._log_dir, ensure_suffix=".log"
        )
        self.__handlers__[config.name].update(config.model_dump(exclude={"name"}))
        return self



from abc import ABC, abstractmethod
from typing import List, Optional

from .base import IAgent
from .pool import IAgentPool


class IAgentSelectionStrategy(ABC):
    """
    智能体选择策略
    """

    def __init__(self, agent_pool: Optional[IAgentPool] = None) -> None:
        super().__init__()
        self.agent_pool = agent_pool

    @abstractmethod
    def select_agent(
        self, agent_pool: Optional[IAgentPool] = None, **kwargs
    ) -> Optional[IAgent]:
        """
        选择符合需求的智能体
        """


class DefaultAgentSelectionStrategy(IAgentSelectionStrategy):
    """
    默认智能体选择策略
    """

    def select_agent(self, agent_pool: Optional[IAgentPool] = None, **kwargs):
        """
        选择所有智能体
        """
        agent_pool = agent_pool or self.agent_pool
        assert agent_pool, "未指定资源池，无法申请资源"
        agent = yield from agent_pool.request_resource(**kwargs)
        return agent



import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, Optional, Type

import simpy
from networkx import DiGraph

from logis.metric import MetricModelType
from logis.task import TaskGraph

from .base import *

if TYPE_CHECKING:
    from ..command.model import SetupArgs
    from ..ctx import Context
    from ..data_type import SimContext


class ISimProxy(ABC):
    """
    仿真模块接口
    """

    @property
    @abstractmethod
    def data_report(self) -> Optional[IDataReport]:
        pass

    @property
    @abstractmethod
    def log(self) -> logging.Logger:
        pass

    @property
    @abstractmethod
    def json_parser(self) -> Optional[IJsonParser]:
        pass

    @property
    @abstractmethod
    def result_generator(self) -> Optional[IResultGenerator]:
        """获取结果生成器（子类需实现）"""
        pass

    @property
    @abstractmethod
    def storage_manager(self) -> Optional[IStorageManager]:
        """获取存储管理器（子类需实现）"""
        pass

    @property
    @abstractmethod
    def network(self) -> DiGraph:
        """获取网络模块（子类需实现）"""
        pass

    @property
    def env(self) -> simpy.Environment:
        assert self.sim_ctx is not None, "sim_ctx is None"
        return self.sim_ctx.env

    @property
    def sim_id(self) -> str:
        return self.context.sim_id()

    @property
    def context(self) -> Type["Context"]:
        from ..ctx import Context

        return Context

    @property
    @abstractmethod
    def task_graph(self) -> TaskGraph:
        """获取任务图（子类需实现）"""
        pass

    @property
    @abstractmethod
    def rack_graph(self) -> DiGraph:
        """获取机架图（子类需实现）"""
        pass

    @property
    @abstractmethod
    def sim_ctx(self) -> Optional["SimContext"]:
        """获取仿真上下文（子类需实现）"""
        pass

    @property
    @abstractmethod
    def debug(self) -> bool:
        """获取调试模式标识（子类需实现）"""
        pass

    @property
    @abstractmethod
    def setup_args(self) -> "SetupArgs":
        """获取启动参数（子类需实现）"""
        pass

    @abstractmethod
    def start_timing_if_not(self, key: str):
        """
        启动计时（若未启动）
        Args:
            key: 计时键名
        """
        pass

    @abstractmethod
    def stop_timing(self, key: str):
        """
        停止计时
        Args:
            key: 计时键名
        """
        pass

    @abstractmethod
    def collect_metric(self, metric: MetricModelType):
        """
        收集指标数据
        Args:
            metric: 指标模型
        """
        pass

    @property
    def logic_graph(self) -> Optional[DiGraph]:
        """基于 json_parser 实现的逻辑图（子类可重写）"""
        x = self.json_parser
        return x.logic_graph if x else None

    @property
    def produce_recipe_graph(self) -> Optional[DiGraph]:
        """基于 json_parser 实现的生产配方图（子类可重写）"""
        x = self.json_parser
        return x._produce_recipe_graph if x else None

    @property
    def id_instance_map(self) -> Optional[Dict[str, Any]]:
        """基于 json_parser 实现的ID实例映射（子类可重写）"""
        x = self.json_parser
        return x.object_map if x else None

    @property
    def is_sim_over(self):
        """判断仿真是否结束（基于 sim_ctx 的默认实现，子类可重写）"""
        ctx = self.sim_ctx
        assert ctx, "仿真上下文未初始化"
        return ctx.is_over()

    @property
    def is_sim_terminated(self):
        """判断仿真是否终止（基于 sim_ctx 的默认实现，子类可重写）"""
        ctx = self.sim_ctx
        assert ctx, "仿真上下文未初始化"
        return ctx.is_terminated()

    @property
    def ws_conn(self):
        """获取WebSocket连接（基于 sim_ctx 的默认实现，子类可重写）"""
        ctx = self.sim_ctx
        return ctx.ws_conn if ctx else None



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional, TypeVar

import simpy

from logis.biz.sim import AgentId
from logis.biz.sim.const import AgentSelectionStrategyName

from .base import AgentClass, IAgent

if TYPE_CHECKING:
    from ..model import AgentManifest, TaskType

T = TypeVar("T")


class IAgentPool(ABC):
    """
    TODO: 继承自ITaskHandler
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self.capacity: Optional[int] = None
        self._store: Optional[simpy.Store] = None

    def cancel_request_resource(self, req: simpy.Event):
        """
        取消申请资源
        """
        self._store.get_queue.remove(req)

    def cancel_release_resource(self, req: simpy.Event):
        """
        取消释放资源
        """
        self._store.put_queue.remove(req)

    def do_request_resource(
        self, **kwargs
    ) -> Generator[simpy.Event, Any, Optional[AgentClass]]:
        """
        执行申请资源
        """
        assert self._store, "未初始化资源池，无法申请资源"
        return self._store.get()

    def do_release_resource(self, resource: AgentClass, *args, **kwargs):
        """
        执行释放资源
        """
        assert self._store, "未初始化资源池，无法释放资源"
        return self._store.put(resource)

    @property
    def resources(self) -> List[IAgent]:
        """
        获取所有的智能体列表
        """
        pass

    @property
    def available_quantity(self):
        return len(self._store.items)

    def get_resource_by_id(self, id: AgentId) -> Optional[AgentClass]:
        """
        根据智能体id获取智能体
        """
        for r in self.resources:
            if r.id == id:
                return r
        return None

    def before_request_resource(self, **kwargs):
        """
        资源申请之前的回调
        """
        raise NotImplementedError()

    @abstractmethod
    def request_resource(
        self,
        *args,
        strategy: Optional[AgentSelectionStrategyName] = None,
        fast_fail: bool = False,
        **kwargs
    ) -> Generator[simpy.Event, Any, Optional[IAgent]]:
        """
        申请资源
        """

    @abstractmethod
    def after_resource_requested(self, **kwargs):
        """
        资源申请到之后的回调
        """
        pass

    def before_release_resource(self, **kwargs):
        """
        资源释放之前的回调
        """
        raise NotImplementedError()

    @abstractmethod
    def release_resource(self, resource: AgentClass, *args, **kwargs):
        """
        释放资源
        """

    def after_resource_released(self, **kwargs):
        """
        资源释放之后的回调
        """
        raise NotImplementedError()

    @abstractmethod
    def assign_resources(
        self, task_type: str, task_keys: List[T], **kwargs
    ) -> Generator[
        simpy.Event, Any, Dict[AgentId, "AgentManifest[AgentClass, TaskType]"]
    ]:
        """
        输入任务分配智能体
        Args:
            task_type: 任务类型
            task_keys: 任务key列表
            kwargs: 其他参数
        Returns:
            智能体id到智能体manifest的映射
        """



from datetime import datetime
from typing import List, Optional

from ipa.decorator import deprecated
from sqlmodel import Field, Index, Relationship, SQLModel, UniqueConstraint


@deprecated("useless for now")
class StockCellMapping(SQLModel, table=False):
    """
    记录货物储位的映射关系，理论上最小单元的货物和储位是多对一的关系，考虑到扩展的可能，这里使用多对多关系
    """

    __tablename__ = "stock_cell_mapping"
    cell_id: Optional[int] = Field(
        foreign_key="cell.id", primary_key=True, default=None
    )
    stock_id: Optional[int] = Field(
        foreign_key="stock.id", primary_key=True, default=None
    )


@deprecated("useless for now")
class Stock(SQLModel, table=False):
    """
    货物信息
    """

    __tablename__ = "stock"

    id: Optional[int] = Field(
        default=None, primary_key=True, sa_column_kwargs={"autoincrement": True}
    )
    sim_id: str
    # 局部唯一或全局唯一ID
    stock_id: str
    name: str
    quantity: Optional[int] = Field(default=None)
    unit: Optional[str] = Field(default=None)
    sim_now: Optional[int] = None
    created_at: datetime = Field(default_factory=datetime.now)
    updated_at: datetime = Field(default_factory=datetime.now)

    # cell_id适用于多对一关系，这里用多对多关系
    cell_id: Optional[str] = None
    # cells适用于多对多关系
    cells: List["Cell"] = Relationship(
        back_populates="stocks", link_model=StockCellMapping
    )

    __table_args__ = (
        # 联合唯一约束
        UniqueConstraint("sim_id", "stock_id"),
        Index("simid_name", "sim_id", "name"),
    )

    def __repr__(self) -> str:
        return f"Stock(id={self.id}, sim_id={self.sim_id}, stock_id={self.stock_id}, name={self.name}, quantity={self.quantity}, unit={self.unit}, sim_now={self.sim_now},  cell_id={self.cell_id})"


@deprecated("useless for now")
class Cell(SQLModel, table=False):
    """
    记录储位信息
    """

    __tablename__ = "cell"

    id: Optional[int] = Field(
        primary_key=True, sa_column_kwargs={"autoincrement": True}, default=None
    )
    sim_id: str
    cell_id: str
    row: Optional[int] = None
    col: Optional[int] = None
    name: Optional[str] = Field(index=True, default=None)
    # id中可能包含rack_id，此时此字段也不必要
    rack_id: str = Field(default="")
    # 虽然多对一关系可以节省空间，但是维护的关系也复杂，目前用不到，先不那么做
    # rack: "Rack" = Relationship(back_populates="cells")
    stocks: List[Stock] = Relationship(
        back_populates="cells", link_model=StockCellMapping
    )

    __table_args__ = (
        # 联合唯一约束
        UniqueConstraint("sim_id", "cell_id"),
    )


@deprecated("useless for now")
class Rack(SQLModel, table=False):
    """
    记录货架信息
    """

    __tablename__ = "rack"

    id: Optional[int] = Field(
        primary_key=True, sa_column_kwargs={"autoincrement": True}, default=None
    )
    sim_id: str = Field(index=True)
    rack_id: str
    name: str = Field(index=True)
    type: Optional[str] = None
    row_count: Optional[int] = None
    col_count: Optional[int] = None

    # cells: List[Cell] = Relationship(back_populates="rack", cascade_delete=True)

    __table_args__ = (
        # 联合唯一约束
        UniqueConstraint("sim_id", "rack_id"),
    )



from abc import abstractmethod
from typing import List, Optional, TypeVar

from logis.biz.sim.task.model import TaskManifest
from logis.iface import IControl
from logis.task import ITaskHandler

from ..data_type.component import BlueprintKind
from .component import IComponent
from .proxy import ISimProxy


class IBlueprint(ISimProxy, ITaskHandler, IComponent, IControl):
    """
    所有蓝图组件的抽象基类
    """

    @abstractmethod
    def get_task_manifest[T](self) -> Optional[TaskManifest[T]]:
        """
        获取任务清单
        """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name: Optional[str] = None
        self.create_edit_id: Optional[str] = None
        self.type_id: Optional[str] = None
        self.type_name: Optional[str] = None

    @classmethod
    def verbose_name(cls, *args, **kwargs):
        return cls.__name__

    @staticmethod
    @abstractmethod
    def kinds(*args, **kwargs) -> List[BlueprintKind]:
        """
        之所以有这个，是因为历史逻辑依靠此标志来分组
        """
        pass


BlueprintClass = TypeVar("BlueprintClass", bound=IBlueprint)



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Union


class IOrder(ABC):

    @abstractmethod
    def get_priority(self) -> int:
        """
        订单优先级,数字越小优先级越高
        """
        pass

    def get_wave_id(self) -> Union[str, int, None]:
        """
        波次ID
        """
        return None

    @abstractmethod
    def set_wave_id(self, wave_id: Union[str, int, None]):
        """
        设置波次ID
        """
        pass

    @abstractmethod
    def get_order_time(self) -> datetime:
        """
        订单时间
        """
        pass



from typing import Literal, Optional, Union

from pydantic import BaseModel


class Notification(BaseModel):
    title: Optional[str] = None
    duration: Union[float, int] = -1
    content: Optional[str] = None
    closable: bool = True
    type: Literal["info", "success", "warning", "error"] = "info"



from typing import Protocol, runtime_checkable


@runtime_checkable
class Producer(Protocol):
    """
    生产者
    """

    def produce(self, *args, **kwargs):
        pass


@runtime_checkable
class Consumer(Protocol):
    """
    消费者
    """

    def consume(self, *args, **kwargs):
        pass



import math
import random
from typing import List, Optional, Tuple

from scipy.spatial.distance import cityblock

from logis.data_type import Point


def manhattan_distance(start_point: Point, end_point: Point) -> float:
    """
    计算曼哈顿距离，又称为街道距离
    """
    return float(cityblock(start_point.to_tuple(), end_point.to_tuple()))
    distance = abs(start_point.x - end_point.x) + abs(start_point.y - end_point.y)
    return distance


def euclid_distance(pos1: Point, pos2: Point, precision: Optional[int] = None) -> float:
    """计算欧几里得距离"""
    v = math.sqrt((pos1.x - pos2.x) ** 2 + (pos1.y - pos2.y) ** 2)
    return round(v, precision) if precision is not None else v


def vector_between_points(p1: Point, p2: Point) -> Tuple[float, float]:
    """计算两点之间的向量。"""
    return p2.x - p1.x, p2.y - p1.y


def angle_between_vectors(v1: Tuple[float, float], v2: Tuple[float, float]) -> float:
    """计算两个向量之间的夹角（弧度）。"""
    dot_product = v1[0] * v2[0] + v1[1] * v2[1]
    magnitude_product = math.sqrt(v1[0] ** 2 + v1[1] ** 2) * math.sqrt(
        v2[0] ** 2 + v2[1] ** 2
    )
    if magnitude_product == 0:
        return 0
    cos_theta = dot_product / magnitude_product
    cos_theta = max(-1, min(1, cos_theta))  # 避免数值误差导致超出 [-1, 1] 范围
    return math.acos(cos_theta)


def intersection_calculate(
    line1_p1: Point, line1_p2: Point, seg_p1: Point, seg_p2: Point
) -> Optional[Point]:
    """计算两条线段的交点"""
    x1, y1 = line1_p1.x, line1_p1.y
    x2, y2 = line1_p2.x, line1_p2.y
    x3, y3 = seg_p1.x, seg_p1.y
    x4, y4 = seg_p2.x, seg_p2.y

    det = (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4)
    if det == 0:
        return None  # 平行或共线

    t = ((x1 - x3) * (y3 - y4) - (y1 - y3) * (x3 - x4)) / det
    u = ((x1 - x3) * (y1 - y2) - (y1 - y3) * (x1 - x2)) / det

    if 0 <= t <= 1 and 0 <= u <= 1:
        intersect_x = x1 + t * (x2 - x1)
        intersect_y = y1 + t * (y2 - y1)
        return Point(intersect_x, intersect_y)

    return None


def _is_point_on_edge(p1: Point, p2: Point, new_pt: Point, eps=1e-2) -> bool:
    """精确判断点是否在边上（含端点）"""

    # 共线性检查
    cross = (p2.x - p1.x) * (new_pt.y - p1.y) - (p2.y - p1.y) * (new_pt.x - p1.x)
    segment_length = math.sqrt((p2.x - p1.x) ** 2 + (p2.y - p1.y) ** 2)
    if segment_length == 0:
        return False  # 线段长度为0
    normalized_cross = abs(cross) / segment_length

    # 判断共线性
    if normalized_cross > eps:
        return False

    # 判断点是否在线段范围内
    dist_p1p2 = segment_length
    dist_p1new = math.sqrt((new_pt.x - p1.x) ** 2 + (new_pt.y - p1.y) ** 2)
    dist_p2new = math.sqrt((new_pt.x - p2.x) ** 2 + (new_pt.y - p2.y) ** 2)
    return abs(dist_p1p2 - (dist_p1new + dist_p2new)) < eps


def uniform_point_in_rectangle(N, center_x, center_y, length, width, rotation_deg=0):
    """
    在矩形内生成均匀分布的点坐标

    参数：
    N (int) - 需要生成的点数
    center_x, center_y (float) - 矩形中心坐标
    length (float) - 矩形长度（旋转前X轴方向）
    width (float) - 矩形宽度（旋转前Y轴方向）
    rotation_deg (float) - 矩形旋转角度（度数，默认0度）

    返回：
    list - 包含N个坐标元组(x, y)的列表
    """
    if N <= 0:
        return []

    # 转换为弧度
    theta = math.radians(rotation_deg)
    cos_theta = math.cos(theta)
    sin_theta = math.sin(theta)

    # 计算行列布局（考虑长宽比优化）
    aspect_ratio = length / width
    cols = max(1, math.ceil(math.sqrt(N * aspect_ratio)))  # 初步列数
    rows = max(1, math.ceil(N / cols))  # 初步行数

    # 优化行列组合（确保足够容纳N个点）
    while rows * cols < N:
        cols += 1
        rows = max(1, math.ceil(N / cols))

    # 计算单元格尺寸
    cell_width = length / cols
    cell_height = width / rows

    # 生成基础网格点（未旋转）
    points = []
    for row in range(rows):
        for col in range(cols):
            # 计算单元格中心坐标（基于未旋转的坐标系）
            x = col * cell_width + cell_width / 2 - length / 2
            y = row * cell_height + cell_height / 2 - width / 2
            points.append((x, y))
            if len(points) >= N:
                break
        if len(points) >= N:
            break

    # 应用旋转和平移变换
    rotated_points = []
    for x, y in points[:N]:
        # 坐标旋转
        x_rot = x * cos_theta - y * sin_theta
        y_rot = x * sin_theta + y * cos_theta

        # 平移至中心点
        x_final = x_rot + center_x
        y_final = y_rot + center_y

        rotated_points.append(Point((round(x_final, 4), round(y_final, 4))))

    return rotated_points


def random_points_in_rectangle(N, center_x, center_y, length, width, rotation_deg=0):
    """在旋转后的矩形内部生成随机分布的坐标点"""
    points = []
    angle_rad = math.radians(rotation_deg)
    cos_theta = math.cos(angle_rad)
    sin_theta = math.sin(angle_rad)

    half_length = length / 2
    half_width = width / 2

    for _ in range(N):
        local_x = random.uniform(-half_length, half_length)
        local_y = random.uniform(-half_width, half_width)

        rotated_x = local_x * cos_theta - local_y * sin_theta
        rotated_y = local_x * sin_theta + local_y * cos_theta

        global_x = center_x + rotated_x
        global_y = center_y + rotated_y

        points.append(Point((global_x, global_y)))

    return points


import math


def vector_length_square(vec: List[float]):
    return sum([x * x for x in vec])


def vector_length(vec: List[float]) -> float:
    return math.sqrt(vector_length_square(vec))


def unit_vector(vec: List[float]) -> List[float]:
    l = vector_length(vec)
    if l == 0:
        raise ZeroDivisionError("zero vector has no unit vector")

    return tuple([x / l for x in vec])


def vector_to(starter: Point, ender: Point) -> Tuple[float, float]:
    return tuple([x - y for (x, y) in zip((ender.x, ender.y), (starter.x, starter.y))])


def vector_add(v1: List[float], v2: List[float]) -> List[float]:
    return tuple([x + y for (x, y) in zip(v1, v2)])


def vector_mul(v: List[float], la: float) -> List[float]:
    return tuple([x * la for x in v])


def weighted_midpoint(starter: Point, ender: Point, la: float):
    """
    gives the point mid so that
    starter-> mid = la * (starter->ender)
    """

    return vector_add(starter, vector_mul(vector_to(starter, ender), la))


def point_on_segment_left_flag(pt, p1, p2):
    """returns pos num if point is on the left side of the segment p1->p2.
    Require points lie on 2d plane."""
    return (p2[0] - p1[0]) * (pt[1] - p1[1]) - (p2[1] - p1[1]) * (pt[0] - p1[0])


def seg_on_seg_same_side(p1, p2, p3, p4):
    """
    returns true if both  p3,p4 lie on the same side of the line p1->p2,
        Require points lie on 2d plane.

    """
    return (
        point_on_segment_left_flag(p1, p3, p4) * point_on_segment_left_flag(p2, p3, p4)
        > 0
    )


def segment_intersection(p1, p2, p3, p4):
    """
    strictly test if intersection happens. Endpoint lie on segment also counts.
    """
    if seg_on_seg_same_side(p1, p2, p3, p4) or seg_on_seg_same_side(p3, p4, p1, p2):
        return False
    else:
        return True


def linear_movement_time(pos1, pos2, velocity):
    assert velocity > 0, f"non-positive velocity {velocity}"

    result = euclid_distance(pos1, pos2) / velocity
    # printpos1, pos2, velocity, result )
    return result


def linear_movement_gen(pos1, pos2, start_t2, total_t1):
    assert total_t1 >= 0, "required to move within negative time"
    if total_t1 == 0:
        assert pos1 == pos2, "required to move within zero time"
        return lambda t: pos1

    def result(check_t2):  ## position at check_time

        ratio = (
            float(check_t2 - start_t2) / total_t1
        )  ## total t must be a number, not timeagain
        assert (
            ratio <= 1.01
        ), f"time exceeds movement requirement: {pos1} to {pos2}, start={start_t2}, time_alotted={total_t1}, at time={check_t2}, or {ratio}*t_a after start time"
        return [a + (b - a) * ratio for (a, b) in zip(pos1, pos2)]

    return result


def linear_stop_gen(pos1, pos2, start_t2, total_t1):
    assert total_t1 >= 0, "required to move within negative time"
    if total_t1 == 0:
        assert pos1 == pos2, "required to move within zero time"
        return lambda t: pos1

    def result(check_t2):

        ratio = (
            float(check_t2 - start_t2) / total_t1
        )  ## total t must be a number, not timeagain
        if ratio <= 1:
            return [a + (b - a) * ratio for (a, b) in zip(pos1, pos2)]
        else:
            return pos2

    return result


def wiz(v, z):
    # given z coordinate, make 2d pt v into a 3d pt.
    return tuple(list(v) + [z])


def wizz(v, height_list):
    # given height list, make a 2d point into a list of 3d pts.
    assert isinstance(height_list, list), f"height_list {height_list} is not a list "
    return [wiz(v, z) for z in height_list]

    ## So wizz result is [] if height_list is empty.


def angle_of_heading(v2):
    # given 2d vector, returns angle of heading in degrees.
    # essentially, it is alpha for exp(i*alpha)
    dx, dy = v2
    assert abs(dx) + abs(dy) >= 0.0001, f"heading vector {v2} almost zero"
    if dx == 0:
        if dy > 0:
            res = math.pi / 2
        else:
            res = 3 * math.pi / 2
    else:
        res = math.atan(dy / dx)
        if dx < 0:
            # either dy >0 or not, II and III quadrant
            res += math.pi
        elif dy < 0:  # and dx >0, IV quadrant
            res += math.pi * 2

    return res * 180 / math.pi



import logging
import sys
from logging import StreamHandler
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler
from pathlib import Path

from colorlog import ColoredFormatter

from .fmt import LoggerFormat


def get_default_console_handler(level=logging.INFO) -> StreamHandler:
    handler = StreamHandler(sys.stdout)
    handler.setFormatter(ColoredFormatter(fmt=LoggerFormat.COLORED_DEFAULT.value))
    handler.setLevel(level)
    return handler



import sys
from typing import Callable, NewType

from logis.alg.path_finding import Finder
from logis.alg.path_finding.model import (
    PathFindingAlgorithmType,
    PathFindingInput,
)
from logis.data_type import Predicate
from logis.data_type.point import Point


def test_pathfinding_model():
    if sys.version_info >= (3, 13):
        assert isinstance(PathFindingAlgorithmType, NewType)
    assert isinstance(Predicate, Callable)

    print("Pathfinding model test passed.")

    input = PathFindingInput(start=Point.of(1, 2), end=Point.of(3, 4))
    input.graph


def test_pathfinding_alg():
    pass



__doc__ = "工具类"


from typing import Any, Callable, Optional, TypeVar

from pandas import isna

from ..data_type import ErrorHandler, cast_if_not_none, round_if_not_none

T = TypeVar("T")


def first_not_none(*args: T) -> Optional[T]:
    for item in args:
        if item is not None:
            return item
    return None


def none_if_in(some_value: Any, *choices: Any):
    """
    如果值在choices中，返回None，否则返回原值
    此方法处理了pd.isna的情况
    """
    if isna(some_value):
        return some_value if next(filter(isna, choices), None) is None else None
    return None if some_value in choices else some_value


def set_attr_if(
    obj: Any,
    props: dict,
    cond: Callable[[str, Any], bool],
    on_error: Optional[ErrorHandler] = None,
    **kwargs,
):
    """
    如果指定对象满足条件则设置目标属性的值

    Args:
        obj: 目标对象
        props: 要设置的属性值
        cond: 过滤器，接收属性名和属性值，返回是否满足条件
        on_error: 错误处理函数，接收异常对象，默认忽略异常
    """
    assert obj is not None, "obj can't be None"
    for k, v in props.items():
        if not cond(k, v):
            continue
        try:
            setattr(obj, k, v)
        except Exception as e:
            if on_error:
                on_error(e)


def set_attr_if_has(obj: Any, props: dict, **kwargs):
    """
    如果目标对象存在指定属性则设置
    """
    set_attr_if(obj, props, lambda k, v: hasattr(obj, k), **kwargs)



def dijkstra(starter, ender, vs, distance_func):
    ## dijkstra, with edge-length-dictionary as input
    ##
    ## WARNING: this design is flawed.
    ## When we attempt to find path for a starter in the middle of an edge, it could be
    ## so near (or virtually at) a graph vertex that the distance is judged to be 0.
    ##
    ##??? FIRST SNAP TO THE GRAPH VERTICES, THEN FIND THE PATH ????

    ## distance_func(v0, v1) = 0 if not connected, = d(v0,v1) otherwise
    ## returns (boolean_find_successful, result )

    openlist = [starter]
    backtrace = dict({})

    # distance from v to source
    dist = dict({starter: 0})

    neighbours = dict({x: [y for y in vs if distance_func(x, y) > 0] for x in vs})

    while openlist != []:

        min_val = None
        min_path = None
        for u in openlist:
            for v in neighbours[u]:
                if v in dist.keys():
                    neighbours[u].remove(v)
                    continue

                res = dist[u] + distance_func(u, v)
                if min_val is None or min_val > res:
                    min_val = res
                    min_path = (u, v)

        if min_val is None:
            ## somehow we don't find any possible expansions and the exploration fails
            break

        else:
            # do have some new min
            u = min_path[0]
            v = min_path[1]
            dist[v] = min_val
            backtrace[v] = u
            openlist.append(v)
            neighbours[u].remove(v)
            if neighbours[u] == []:
                openlist.remove(u)

            if v == ender:
                return (
                    True,
                    # trace_to(ender, starter, backtrace),
                    dist[v],
                )
            # by default gives [starter,...ender]

    return (False, None)



from typing import Optional

from pydantic import AliasChoices, ConfigDict
from sqlmodel import Field, SQLModel

from logis.util.pydantic_util import collect_field_aliases


class TestTableModel(SQLModel, table=False):
    model_config = ConfigDict(strict=False, coerce_numbers_to_str=True)
    name: str = Field(
        "",
        alias="名字",
        schema_extra=dict(
            validation_alias=AliasChoices("userName", "name"), strict=False
        ),
    )

    age: Optional[int] = Field(default=None)


def test_collect_alias():
    name_field = TestTableModel.model_fields["name"]
    aliases = collect_field_aliases(name_field)

    assert not aliases.difference(["名字", "userName", "name"])


def test_implict_type_cast():
    m = TestTableModel.model_validate(dict(age="10", name=100))
    assert m.age == 10
    assert m.name == "100"



import json

from pyecharts.charts import Bar


def test_bar():
    bar = Bar()
    bar.add_xaxis(["A", "B", "C", "D", "E"])
    bar.add_yaxis("Series 1", [1, 2, 3, 4, 5])

    x = bar.dump_options()
    y = json.loads(x)

    x = bar.dump_options_with_quotes()
    y = json.loads(x)



__doc__ = """
仿真模块接口
"""


from .base import *
from .blueprint import *
from .component import *
from .order import *
from .picking import *
from .proxy import *
from .registry import *
from .wave import *



import heapq
import math
from collections import defaultdict
from typing import Callable, Dict, List, Optional, Set, Tuple

from logis.data_type import Point
from logis.math import euclid_distance

from .base import A_STAR_ALGORITHM, PathFindingAlgorithm, track_path
from .model import PathFindingInput, PathFindingOutput


def a_star(
    starter: Point,
    ender: Point,
    nb_dis_dic: Dict[
        Point, List[Tuple[Point, float]]
    ],  # 每个点的邻居点和距离和所需要的速度以及该点的角速度
    heuristic_func: Callable[[Point, Point], float] = None,
    excluded_vertices: List[Point] = [],
    input: Optional[PathFindingInput] = None,
):
    """
    在图中使用 A* 算法寻找从起始点到终点的路径。

    参数:
        starter (Point): 起始点。
        ender (Point): 终点。
        nb_dis_dic (Dict[Point, List[Tuple[Point, float, float, float]]]):
            字典，键为顶点，值为元组列表，元组包含 (邻居顶点, 距离, 速度, 角速度)。
        heuristic_func (Callable[[Point, Point], float], 可选):
            启发式函数，接受两个顶点并返回启发式距离。默认为欧几里得距离。
        excluded_vertices (List[Point], 可选): 要排除的顶点列表。默认为空列表。

    返回:
        Tuple[bool, List[Point], List[float], List[float]]:
            若找到路径，返回 (True, 路径列表, 速度列表, 转向时间列表)；
            若未找到路径，返回 (False, None, [], [])。
    """

    if starter == ender:
        return PathFindingOutput(success=True, paths=[[starter]])

    # If no heuristic function is provided, use Euclidean distance as default
    if heuristic_func is None:
        heuristic_func = euclid_distance

    # Priority queue to store the vertices to be explored
    frontier: List[Tuple[float, float, Point]] = []
    heapq.heappush(frontier, (0, 0, starter))  # (current_cost, heuristic_cost, vertex)

    came_from: Dict[Point, List[Point]] = defaultdict(list)
    # Dictionary to store the cost of getting to each vertex
    cost_so_far = {starter: 0}
    # 用于记录到达每个顶点所经过边的速度
    speed_so_far = {starter: None}
    # 用于记录到达每个顶点所经过边的角速度
    angular_speed_so_far = {starter: None}
    # 用于记录到达每个顶点时的前一个顶点，用于计算转向
    prev_vertex = {starter: None}
    # Set to keep track of visited vertices
    visited: Set[Point] = set()
    success = False
    while frontier:
        _, _, current = heapq.heappop(frontier)

        if current == ender:
            success = True
            break

        visited.add(current)

        for neighbor, weight in nb_dis_dic[current]:
            if neighbor in excluded_vertices:
                continue
            # 如果节点已经被处理过就不再处理
            if neighbor in visited:
                continue
            # TODO:如果节点已经在待处理队列中就跳过?
            if neighbor in [vertex[2] for vertex in frontier]:
                continue
            new_cost = cost_so_far[current] + weight
            # 如果新路径更长就跳过
            old_cost = cost_so_far.get(neighbor, math.inf)
            if new_cost > old_cost:
                continue

            # 注释掉下面的判断可以生成多条路径
            # # 如果达到neighbor的成本相同，就只选择第一次选中的那个，从而避免产生多条等长的路径
            if new_cost == old_cost:
                continue

            if new_cost < old_cost:
                came_from[neighbor].clear()

            cost_so_far[neighbor] = new_cost
            priority = new_cost + heuristic_func(neighbor, ender)
            heapq.heappush(frontier, (priority, new_cost, neighbor))
            came_from[neighbor].append(current)

    # 创建正向连接关系方便后续遍历
    step_into: Dict[Point, List[Point]] = defaultdict(list)
    for target, sources in came_from.items():
        for source in sources:
            step_into[source].append(target)

    # 如果找到了起点到终点的路径就剪枝回溯
    if success:
        assert ender in came_from, "寻路算法出错了"
        step_into.clear()
        targets = [ender]
        while targets:
            target = targets.pop()
            sources = came_from.get(target, [])
            for source in sources:
                step_into[source].append(target)
            targets.extend(sources)

    if not success:
        return PathFindingOutput(success=False, paths=[])

    # 如果没找到直接到终点的路线，就尝试找一个最接近终点的点
    paths = track_path(step_into, starter, ender, limit=None if success else 1)
    paths = list(
        filter(lambda p: len(p) > (1 if p[0] in [starter, ender] else 0), paths)
    )
    return PathFindingOutput(success=success, paths=paths)


class AStarPathFinding(PathFindingAlgorithm):
    """
    A*寻路算法实现
    """

    type = A_STAR_ALGORITHM

    def find_path(self, input: PathFindingInput, **kwargs) -> PathFindingOutput:
        """
        使用A*算法寻找从start到end的路径
        """
        return a_star(
            input=input,
            starter=input.start,
            ender=input.end,
            nb_dis_dic=input.distance_matrix,
            heuristic_func=input.heuristic_func,
            excluded_vertices=set(input.excluded_vertices),
            **kwargs,
        )



from abc import ABC, ABCMeta
from collections import defaultdict
from typing import Dict, Generic, Protocol, TypeVar, Union, runtime_checkable

from logis.data_type import NumberType, NumberUnit, SpatialProps
from logis.util.num_util import get_numeric_value


class Shape(metaclass=ABCMeta):
    """
    形状，例如路径、节点等
    """

    @property
    def rotation_value(self):
        v = get_numeric_value(self.props.rotation) if self.props else None
        return v or 0

    def __init__(self, **kwargs):
        self.props: Union[SpatialProps, None] = None


from ._event import *
from ._queue import *
from .container import *
from .graph import *
from .manage import *
from .registry import *



import logging
from typing import Any, Dict, Optional

from logis.data_type import TmpId


class ComponentForm:
    """
    entity模块，用来存储交互文件中结点（事件、资源池、货架组等）以及所有空间标记的信息

    TODO:类型标注、泛型
    """

    def __init__(self, entity_data: Dict[str, Any], **kwargs):
        self.type: str = entity_data.get("Type")
        # 实际上是类型id
        self.type_id = self.id = str(entity_data.get("ID"))
        self.type_name: str = entity_data.get("Name", "Unknown")
        self.create_edit_id: str = str(entity_data.get("Create_Edit_ID", ""))
        # TODO：这个字段看起来没什么用，考虑去除
        self.agent_id: TmpId = entity_data.get("AgentID")
        # TODO：这个字段还没看到在哪用的
        self.link_ids: Dict[str, Any] = entity_data.get("LinkIDs", {})
        self.extends_dict: Dict[str, Any] = {}
        self.properties: Dict[str, Any] = {}
        self.disabled: bool = entity_data.get("disabled", False)

        if not self.create_edit_id:
            logging.warning(
                f"Warning: Entity {self.type_name} is missing 'Create_Edit_ID'"
            )

    @property
    def name(self) -> Optional[str]:
        return self.properties.get("名称")

    @property
    def display_name(self) -> Optional[str]:
        return self.properties.get("显示名称")



import logging
import math
import time
from typing import Any, List, Optional, Union

import simpy
from simpy.core import EmptySchedule, StopSimulation
from simpy.events import URGENT

from .exception import DeadlineException


def try_put(
    env: simpy.Environment,
    container: simpy.Container,
    amount: Union[int, float],
    timeout: Optional[int] = None,
):
    x = container.put(amount)
    if timeout:
        x = x | env.timeout(timeout)
    v = yield x
    return v


def get_free_capacity(container: simpy.Container):
    """
    获取容器的空闲容量
    Args:
        container (simpy.Container): 容器
    Returns:
        int | float: 空闲容量
    """
    # TODO：这里预先触发存取操作，可能会有性能影响，后续可考虑分离出去
    container._trigger_put(None)
    container._trigger_get(None)
    return container.capacity - container.level


def resize_container(
    container: simpy.Container,
    capacity: Union[int, float],
    is_delta=False,
    force: bool = False,
):
    """
    调整容器容量
    Args:
        container (simpy.Container): 容器
        capacity (int | float): 新容量
        is_delta (bool, optional): 是否是增量调整. 默认False
        force (bool, optional): 是否强制调整. 默认False
    """
    old_capacity = container.capacity
    capacity = capacity + old_capacity if is_delta else capacity

    if force:
        container._capacity = capacity
        return container.capacity

    dx = capacity - old_capacity
    if not dx:
        return container.capacity
    # 如果是减容，先触发取操作
    if dx < 0:
        container._trigger_get(None)
        assert (
            container.level <= capacity
        ), f"容器当前等级{container.level}大于新容量{capacity}，无法完成减容"
    container._capacity = capacity
    if dx > 0:
        container._trigger_put(None)

    return container.capacity


def resize_container_level(
    container: simpy.Container,
    level: Union[int, float],
    is_delta=False,
    force: bool = False,
):
    """
    调整容器库存
    Args:
        container (simpy.Container): 容器
        level (int | float): 新库存
        is_delta (bool, optional): 是否是增量调整. 默认False
        force (bool, optional): 是否强制调整. 默认False
    """
    old_level = container.level
    level = level + old_level if is_delta else level

    if force:
        container._level = level
        return container.level

    dx = level - old_level
    if not dx:
        return container.level
    # 如果是缩小库存
    if dx < 0:
        container._trigger_put(None)
        assert container.level >= abs(
            dx
        ), f"容器当前库存{container.level}无法缩减{dx}至{level}"
    else:
        container._trigger_get(None)
        assert (
            level <= container.capacity
        ), f"容器当前库存{container.level}无法增加{dx}至{level}，因为最大容量为{container.capacity}"
    container._level = level

    return container.level


def resource_to_dict(resource: Any):
    if isinstance(resource, simpy.Container):
        return dict(level=resource.level, capacity=resource.capacity)
    if isinstance(resource, simpy.Resource):
        return dict(level=resource.count, capacity=resource.capacity)
    if isinstance(resource, simpy.Store):
        return dict(level=len(resource.items), capacity=resource.capacity)
    logging.error("unexpected type:" + resource)


def schedule_event_at(
    env: simpy.Environment, at: Union[float, int], ev: Optional[simpy.Event] = None
):
    """
    计划在指定时间触发事件
    Args:
        env (simpy.Environment): 仿真环境
        at (float | int): 触发时间
        ev (simpy.Event | None, optional): 事件. 默认None
    """
    if at < env.now:
        return
    if ev is None:
        ev = env.event()
        ev._ok = True
        ev._value = None
    env.schedule(ev, URGENT, at - env.now)
    return ev


def interrupt_on_event(env: simpy.Environment, ev: simpy.Event):
    """
    当事件触发时中断仿真
    Args:
        env (simpy.Environment): 仿真环境
        ev (simpy.Event): 事件
    """

    raise NotImplementedError("interrupt_on_event not implemented")


def interrupt_if_timeout(
    env: simpy.Environment,
    timeout: Optional[float] = None,
    check_interval: float = 10,
    exit_signal: Optional[simpy.Event] = None,
):
    """
    根据仿真真实运行时间是否超时来中断仿真

    Args:
        env (simpy.Environment): 仿真环境
        timeout (float | None, optional): 最大运行时间. 默认不限时
        check_interval (float, optional): 这个时间是仿真中的时间
        exit_signal (simpy.Event | None, optional): 外部信号，用于外部中断仿真. 默认不使用.
    """

    def inner():
        cpu_start_time = time.time()
        while True:
            dt = time.time() - cpu_start_time
            if timeout is not None and dt > timeout:
                raise simpy.Interrupt(
                    f"仿真运行时间{dt}超过最大运行时间{timeout}，强制退出"
                )
            if exit_signal and exit_signal.triggered:
                raise simpy.Interrupt(f"外部信号{exit_signal}触发，强制退出")
            yield env.timeout(check_interval)

    return env.process(inner())


def has_no_event_left(env: simpy.Environment):
    """
    判断指定环境中是否还有未处理的事件
    """
    return not env._queue and not env.active_process


def _max_run_time(env: simpy.Environment, v: Optional[float] = None):
    key = "__max_run_time__"
    if v is not None:
        setattr(env, key, v)
    return getattr(env, key, None)


def _max_sim_time(env: simpy.Environment, v: Optional[float] = None):
    key = "__max_sim_time__"
    if v is not None:
        setattr(env, key, v)
    return getattr(env, key, None)


def _started_at(env: simpy.Environment, v: Optional[float] = None):
    key = "__started_at__"
    if v is not None:
        setattr(env, key, v)
    return getattr(env, key, None)


def run_until(
    env: simpy.Environment,
    exit_signal: Optional[simpy.Event] = None,
    max_sim_time: Optional[float] = None,
    extra_events: Optional[List[simpy.Event]] = None,
    max_run_time: Optional[float] = None,
    **kwargs,
):
    """

    运行仿真直到指定事件触发或仿真时间超过最大运行时间

    TODO: 如果重复调用,可能添加多次until事件,如何规避?

    Args:
        env (simpy.Environment): 仿真环境
        exit_signal (simpy.Event | None, optional): 外部信号，用于外部中断仿真. 默认不使用.
        max_sim_time (float | None, optional): 最大运行时间. 默认不限时
        extra_events (List[simpy.Event] | None, optional): 额外事件. 默认不使用.
    """
    log: logging.Logger = kwargs.get("log", logging)
    if getattr(env, "__old_step__", None) is None:
        env.__old_step__ = env.step

    if _started_at(env) is None:
        _started_at(env, time.time())
    if max_sim_time is not None:
        _max_sim_time(env, max_sim_time)
    if max_run_time is not None:
        _max_run_time(env, max_run_time)

    started_at = _started_at(env)
    max_run_time = _max_run_time(env)
    max_sim_time = _max_sim_time(env)

    def _step():
        next_time = env.peek()
        if next_time is None or next_time == math.inf:
            raise EmptySchedule from None
        if max_sim_time is not None and next_time > max_sim_time:
            log.warning(
                "仿真时长已超限制：%s，即将停止仿真",
                max_sim_time,
            )
            raise DeadlineException(
                f"next simulation time {next_time} will exceeds max_sim_time {max_sim_time}"
            )
        if (
            max_run_time is not None
            and started_at is not None
            and time.time() - started_at > max_run_time
        ):
            log.warning(
                "运行时长已超限制：%s，即将停止仿真",
                max_run_time,
            )
            raise DeadlineException(
                f"next simulation time {next_time} will exceeds max_run_time {max_run_time}"
            )
        env.__old_step__()

    events = []
    if exit_signal:
        events.append(exit_signal)

    # 这种方式会将事件加入调度队列,还是会造成仿真延迟
    #     # dt = max(0, max_sim_time - env.now)
    #     # events.append(env.timeout(dt))
    #     events.append(DeadlineEvent(env, max_sim_time))
    if extra_events:
        events.extend(extra_events)
    until = env.any_of(events) if events else None
    try:
        env.step = _step
        env.run(until=until)
        if exit_signal and exit_signal.triggered:
            log.info("%s收到终止仿真信号，仿真退出", env.now)
    except Exception as e:
        if has_no_event_left(env):
            log.info("error happens but there is no event left, will ignore: %s", e)
            return
        raise e


def stop_simulation(env: simpy.Environment):
    """
    停止仿真
    1. 清空事件队列，诱发底层异常
    """

    def clear_queue():
        size = len(env._queue)
        logging.debug("will remove %s events in queue", size)
        env._queue.clear()
        yield env.timeout(0)
        return size

    return env.process(clear_queue())



from logis.data_type import Point
from logis.iface import IEdge
from logis.math import euclid_distance, manhattan_distance


class DirEdge(IEdge):
    """
    有向图的连边

    注意：此结构体已废弃，后续统一使用networkx相关
    """

    def __init__(self, starter: Point, ender: Point, speed: float = -1, **kwargs):
        self.starter = starter
        self.ender = ender
        # TODO:之前一直用的是欧几里得距离，在仓储场景下，好像曼哈顿距离更合适
        # self.length = manhattan_distance(starter, ender)
        self.length = euclid_distance(starter, ender)
        self.speed = speed

    def get_weight(self, *args, **kwargs) -> float:
        return self.length

    def is_directed(self) -> bool:
        return True

    def __str__(self):
        return f"e({self.starter}-{self.ender})"

    def __repr__(self):
        return str(self)



from abc import ABC, abstractmethod
from datetime import timedelta
from typing import List, Optional
from uuid import uuid4

import pandas
from pydantic import BaseModel

from logis.data_type import DEFAULT_PYDANTIC_MODEL_CONFIG

from .order import *


class WaveGroupingConfig(BaseModel):
    """
    波次划分配置
    """

    # 批次大小
    batch_size: Optional[int] = None
    # 时间窗口
    time_interval: Optional[timedelta] = None
    # 指定各个时间段
    time_index: Optional[pandas.DatetimeIndex] = None

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG


class WaveGroupingStrategy(ABC):
    """
    波次划分策略
    """

    @abstractmethod
    def group(
        self, orders: List[IOrder], config: WaveGroupingConfig
    ) -> List[List[IOrder]]:
        """
        对订单列表进行波次划分
        Args:
            orders: 订单列表
            config: 波次拣选配置
        Returns:
            波次列表
        """
        pass

    def _generate_wave_id(self, waves: List[List[IOrder]]) -> Optional[Union[str, int]]:
        """
        生成波次ID,默认是uuid4
        """
        if not waves:
            return None
        for i, wave in enumerate(waves):
            wave_id = str(uuid4())
            for od in wave:
                od.set_wave_id(wave_id)


class TimeWindowWaveStrategy(WaveGroupingStrategy):
    """
    基于时间窗口的波次划分策略
    """

    def group(self, orders, config):
        if not orders:
            return []
        orders = sorted(orders, key=lambda od: od.get_order_time())

        min_time, left = orders[0].get_order_time(), orders[-1].get_order_time()
        time_index = config.time_index or pandas.date_range(
            min_time, left, freq=config.time_interval
        )
        waves: List[List[IOrder]] = []
        count = 0
        for i, left in enumerate(time_index):
            right = time_index[i + 1] if i + 1 < len(time_index) else None
            wave = []
            for od in orders[count:]:
                if od.get_order_time() >= left and (
                    right is None or od.get_order_time() < right
                ):
                    wave.append(od)
                else:
                    break
            if wave:
                waves.append(wave)
                count += len(wave)

        self._generate_wave_id(waves)
        return waves


class WaveSizeStrategy(WaveGroupingStrategy):
    """
    基于波次大小的波次划分策略
    """

    def group(self, orders, config):
        if not orders:
            return []
        orders = sorted(orders, key=lambda od: od.get_order_time())
        waves: List[List[IOrder]] = []
        wave = []
        for od in orders:
            if len(wave) >= config.batch_size:
                waves.append(wave)
                wave = []
            wave.append(od)
        if wave:
            waves.append(wave)
        self._generate_wave_id(waves)
        return waves



from numbers import Number
from typing import Any, Dict, List, Optional, Union

from pydantic import AliasChoices, BaseModel, Field

from logis.biz.sim.stock import QuantifiedStock
from logis.data_type import (
    DEFAULT_PYDANTIC_MODEL_CONFIG,
    QuantifiedValue,
    Time,
)


class RecipeItemVo(QuantifiedStock):
    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    name: str = Field(validation_alias=AliasChoices("FreightType", "name"))
    quantity: Union[int, Number] = Field(
        validation_alias=AliasChoices("FreightCount", "quantity"), default=1
    )
    unit: str = Field(validation_alias=AliasChoices("FreightUnit", "unit"))

    current_location: Any = None
    target_location: Any = None

    @property
    def unique_id(self):
        return f"{self.name}"

    def __str__(self):
        return f"{self.name}({self.quantity}{self.unit})"


Material = RecipeItemVo
Product = RecipeItemVo


class Recipe(BaseModel):
    """
    配方：描述什么材料生成什么产品，多对多关系，应作为最小单元使用
    """

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    materials: List[RecipeItemVo]
    products: List[RecipeItemVo]

    # 耗时
    cycle_time: Time

    def material_dict(self):
        dc: dict[str, QuantifiedValue] = dict()
        for m in self.materials:
            k = m.unique_id
            dc[k] = m
        return dc

    def product_dict(self):
        dc: Dict[str, RecipeItemVo] = dict()
        for p in self.products:
            k = p.unique_id
            dc[k] = p
        return dc

    def io_dict(self):
        """
        获取材料和产品的映射关系
        """
        return dict(**self.material_dict(), **self.product_dict())

    def get_unit(self, name: str):
        return self.io_dict().get(name).unit


class ProduceProperties(BaseModel):
    """
    生产设备属性
    """

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    name: Optional[str] = Field(validation_alias=AliasChoices("名称"))

    # 基本配方
    recipe: Optional[Recipe] = None
    # 原料
    materials: List[Material] = []
    # 产出
    products: List[Product] = []
    # 有时候可能只需要知道数量
    product_quantity: Optional[Number] = None



__doc__ = """
指标模块
"""


from abc import ABC, abstractmethod
from typing import List, Optional, TypeVar

from logis.util.io_util import AsyncWriteBuffer, WriteBuffer, WriteBufferConfig
from logis.util.lambda_util import invoke

T = TypeVar("T")

class MetricModel(ABC):
    """
    指标模型
    """

    @classmethod
    @abstractmethod
    def get_measurement(
        cls, prefix: Optional[str] = None, suffix: Optional[str] = None, sep: str = ""
    ) -> str:
        """
        获取指标名称
        """
        pass

    def to_influxdb_point(self, measurement: Optional[str] = None):
        """
        转换为InfluxDB点
        """
        raise NotImplementedError("to_influxdb_point")


MetricModelType = TypeVar("MetricModelType", bound=MetricModel)


class IMetricCollector(ABC):
    """
    指标收集器
    """

    def __init__(self, write_buffer: Optional[WriteBuffer] = None, **kwargs):
        super().__init__(**kwargs)
        self._buffer = write_buffer or AsyncWriteBuffer(
            WriteBufferConfig(batch_size=100, flush_interval=10, handler=self.submit)
        )
        invoke(self._buffer.start)

    @abstractmethod
    def submit(self, metrics: List[T]):
        """
        提交指标
        """
        pass

    def collect(self, metric: T):
        """
        缓冲指标
        """
        invoke(self._buffer.put, metric)

    def flush(self, flush_all: bool = False, **kwargs):
        """
        刷新指标
        """
        invoke(self._buffer.flush, flush_all=flush_all, **kwargs)



from abc import ABCMeta
from collections import defaultdict
from typing import Dict, Protocol, runtime_checkable

from logis.data_type import NumberType, NumberUnit


class StoreResult(metaclass=ABCMeta):
    """
    存储结果
    """

    pass


class RetrieveResult(metaclass=ABCMeta):
    """
    检索结果
    """

    pass


class StoreStrategy(metaclass=ABCMeta):
    """
    存储策略
    """


@runtime_checkable
class Storable(Protocol):
    # class Storable(metaclass=ABCMeta):
    """
    可存储的

    使用鸭子类型实现，相比ABC更灵活
    """

    def pre_store(self, *args, **kwargs):
        """
        预存储，一般用来实现资源检查、校验
        """
        pass

    def pre_retrieve(self, *args, **kwargs):
        """
        预取回，一般用来实现资源检查、校验
        """
        pass

    def store(self, *args, **kwargs) -> StoreResult:
        """
        真正的存储操作
        """
        pass

    def retrieve(self, *args, **kwargs) -> RetrieveResult:
        """
        真正的取回操作
        """
        pass


class QuantifiedValueContainer(Storable):

    def __init__(self):
        self.__container__: Dict[str, Dict[str, NumberType]] = defaultdict(
            defaultdict(0)
        )

    def store(self, v: "NumberUnit"):
        self.__container__[v.kind][v.unit] += v.quantity

    def retrieve(self, v: "NumberUnit"):
        self.__container__[v.kind][v.unit] -= v.quantity



import math

from logis.data_type.unitable import DEFAULT_UNIT_CONFIG
from logis.data_type.unitable import Quantity as Q_
from logis.data_type.unitable import get_unit_ratio


def test_unit_config():
    unit_config = DEFAULT_UNIT_CONFIG
    assert unit_config.get_ratio("m", "cm", 10) == 1000
    assert unit_config.get_ratio("km", "cm") == 1000 * 100


def test_unit():
    speed = Q_(60, "千米") / Q_(1, "小时")
    assert math.floor(speed.to("米/秒").magnitude) == 16

    assert Q_(36000, "千米/小时").to("千米/秒").magnitude == 10
    assert Q_(36000, "km/h").to("km/s").magnitude == 10

    assert get_unit_ratio("千米每小时", "千米/秒", 36000) == 10

    assert int(get_unit_ratio("年", "天")) == 365
    assert int(get_unit_ratio("年", "月")) == 12
    assert int(get_unit_ratio("月", "天")) == 30
    assert get_unit_ratio("天", "小时") == 24
    assert get_unit_ratio("星期", "天") == 7
    assert get_unit_ratio("周", "天") == 7

    assert get_unit_ratio("吨", "千克") == 1000

    assert get_unit_ratio("千米", "米") == 1000

    assert round(get_unit_ratio("摄氏度", "华氏度"), 1) == 1.8



from logis.ctx import Context


def test_counter():
    assert next(Context.counter()) == 0
    assert next(Context.counter()) == 1
    assert next(Context.counter(id="1")) == 0
    assert next(Context.counter(id="1")) == 1
    assert next(Context.counter(id="2", start=1)) == 1
    assert next(Context.counter(id="3", start=1, step=2)) == 1
    assert next(Context.counter(id="3")) == 3



import json
import logging
import platform
import re
import subprocess
from pathlib import Path
from typing import Optional, Union

import requests
from influxdb_client_3 import Point as _Point
from pydantic import BaseModel

from logis.data_type import DEFAULT_PYDANTIC_MODEL_CONFIG, ApiResponse


class Point(_Point):
    """
    继承自 influxdb_client_3.Point，避免用法上的错误
    """

    def tag(self, key: str, value: Optional[str] = None):
        """
        避免value为None，确保value一定是字符串类型
        """
        if value is not None:
            super().tag(key, str(value))
        return self


class InfluxQuery(BaseModel):
    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG
    query: str
    language: str = "sql"
    mode: str = "all"
    database: Optional[str] = None


class InfluxCommand:

    def __init__(
        self,
        binary_dir: Path,
        version="v3",
        token: Optional[str] = None,
        host: str = "127.0.0.1",
        port: int = 8181,
    ):
        self._binary_dir = binary_dir
        self.version = version
        self._token = token
        self.port = port
        self.host = host

    @property
    def host_url(self):
        return f"http://{self.host}:{self.port}"

    @property
    def token(self):
        return self._token

    @property
    def binary(self):
        p = self._binary_dir / "influxdb3.exe"
        assert p.exists(), f"InfluxDB 3 binary not found: {p}"
        return p

    def serve(
        self,
        data_dir: Union[Path, str],
        token_file: Optional[Path] = None,
        node_id: Optional[str] = platform.node(),
        object_store: str = "file",
        start_new_session: bool = False,
        creationflags=subprocess.CREATE_NO_WINDOW,
        **kwargs,
    ):
        """
        启动 InfluxDB 服务。
        """
        p = subprocess.Popen(
            list(
                filter(
                    lambda x: x,
                    [
                        str(self.binary),
                        "serve",
                        f"--http-bind={self.host}:{self.port}",
                        f"--node-id={node_id}" if node_id else "",
                        f"--object-store={object_store}" if object_store else "",
                        f"--data-dir={data_dir}" if data_dir else "",
                        f"--admin-token-file={token_file}" if token_file else "",
                    ],
                )
            ),
            stdout=subprocess.PIPE,  # 捕获标准输出
            stderr=subprocess.PIPE,  # 捕获标准错误
            text=True,
            start_new_session=start_new_session,
            creationflags=creationflags,
            **kwargs,
        )
        return p

    def generate_admin_token(
        self,
        name: Optional[str] = None,
        offline: bool = True,
        path: Optional[Path] = None,
        expiry: Optional[str] = None,  # "100y",
        creationflags=subprocess.CREATE_NO_WINDOW,
    ):
        if offline:
            assert (
                path is not None
            ), "Path must be provided for offline token generation"

        with subprocess.Popen(
            list(
                filter(
                    lambda x: x,
                    [
                        str(self.binary),
                        "create",
                        "token",
                        "--admin",
                        f"--name={name}" if name else "",
                        "--offline" if offline else "",
                        f"--expiry={expiry}" if expiry else "",
                        f"--output-file={path}" if offline else "",
                        f"--host={self.host_url}",
                    ],
                )
            ),
            stdout=subprocess.PIPE,  # 捕获标准输出
            stderr=subprocess.PIPE,  # 捕获标准错误
            text=True,
            creationflags=creationflags,
        ) as p:
            stdout, stderr = p.communicate()
        if p.returncode != 0:
            return None
        stdout = re.sub(r"\x1b\[\d+m", "", stdout)
        lines = stdout.split("\n")
        prefix = "Token: "
        for line in lines:
            if line.startswith(prefix):
                return line.split(prefix)[1].strip()
        return None

    def execute(self, cmd: str, token: Optional[str] = None, **kwargs):
        """
        执行 InfluxDB 命令。
        """
        with subprocess.Popen(
            list(
                filter(
                    lambda x: x,
                    [
                        str(self.binary),
                        *cmd.split(" "),
                        f"--token={token}" if token else "",
                        f"--host={self.host_url}",
                    ],
                )
            ),
            stdout=subprocess.PIPE,  # 捕获标准输出
            stderr=subprocess.PIPE,  # 捕获标准错误
            text=True,
            creationflags=subprocess.CREATE_NO_WINDOW | subprocess.DETACHED_PROCESS,
        ) as p:
            stdout, stderr = p.communicate()
        p.wait()
        if p.returncode != 0:
            raise RuntimeError(f"Command failed: {cmd}\n{stderr}")
        return stdout

    def create_database(
        self, database: str, token: str, retention_period: Optional[str] = None
    ):
        """
        创建一个新的数据库。
        """
        rp = f"--retention-period {retention_period}" if retention_period else ""
        return self.execute(
            f"CREATE DATABASE {database} {rp}",
            token=token,
        )

    def list_database(self, token: str):
        """
        列出所有数据库。
        """
        items = json.loads(self.execute("show databases --format json", token=token))

        for item in items:
            item["name"] = item.pop("iox::database")

        return items

class InfluxRestClient:
    """
    官方SDK提供的接口太少了，这里暂且自己封装下
    """

    def __init__(self, base_url: str, token: str, version="v3", **kwargs):
        self._url = base_url.strip().rstrip("/") + "/api/" + version
        self._token = token

    def _get_default_headers(self):
        return dict(Authorization=f"Bearer {self._token}")

    def create_database(self, database: str, strict: bool = True):
        r = requests.post(
            f"{self._url}/configure/database",
            json=dict(db=database),
            headers=self._get_default_headers(),
        )
        r = ApiResponse.from_http_response(r)
        if not strict and r.error and r.error == 409:
            logging.warning(f"Database {database} already exists")
            r.success = True
        return r

    def delete_database(self, database: str, hard_delete_at: Optional[str] = None):
        r = requests.delete(
            f"{self._url}/configure/database",
            params=dict(db=database, hard_delete_at=hard_delete_at),
            headers=self._get_default_headers(),
        )
        return ApiResponse.from_http_response(r, content_type="json")

    def delete_table(
        self,
        database: str,
        table: str,
        hard_delete_at: Optional[str] = None,
        strict: bool = True,
    ):
        """
        Args:
            database: 数据库名称
            table: 表名称
            hard_delete_at: 强制删除时间，格式为RFC3339
            strict: 是否严格模式，默认True, 若设置为False, 则当表不存在时不会报错
        """
        tmp = requests.delete(
            f"{self._url}/configure/table",
            params=dict(db=database, table=table, hard_delete_at=hard_delete_at),
            headers=self._get_default_headers(),
        )
        r = ApiResponse.from_http_response(tmp, content_type="json")
        if not strict and r.error and r.error.code == 404:
            r.error = None
            r.success = True
        return r



from pathlib import Path
from typing import Literal, Optional

from pydantic import BaseModel

from logis.data_type.base import DEFAULT_PYDANTIC_MODEL_CONFIG


class SetupArgs(BaseModel):

    # 是否将其他智能体视为障碍物
    agent_as_obstacle: Optional[bool] = None
    # 是否将货架视为障碍物
    rack_as_obstacle: Optional[bool] = None

    # 是否在退出任务前校验货物量是否一致
    disable_verify_stock_quantity_before_exit: bool = False

    # 拣选货架是否开启储位
    enable_picking_rack_cell: bool = False

    # 是否打印版本号
    version: bool = False

    max_sim_time: Optional[float] = None
    max_run_time: Optional[float] = None

    server_version: Literal["v1", "v2"] = "v2"
    host: str = "127.0.0.1"
    port: int = 8765

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    disable_metric: bool = False
    influxdb_config: Optional[Path] = None

    filter_order_type: Optional[str] = None
    filter_order_offset: Optional[int] = None
    filter_order_limit: Optional[int] = None

    echo: bool = False
    with_memory_db: bool = False
    db_url: Optional[str] = None
    enable_decorators: bool = False
    model: Optional[str] = None
    model_path: Optional[Path] = None
    level: Optional[str] = "WARNING"
    with_node_listener: bool = False
    enable_mapf: bool = False
    with_model_picker: bool = False

    debug: bool = False

    def parse_db_url(self) -> Optional[str]:
        if self.with_memory_db:
            return "sqlite:///:memory:"
        return self.db_url



import logging
import types
from collections import defaultdict
from math import floor
from typing import Any, Generator, List, Optional, Protocol, runtime_checkable

from logis.data_type import merge_quantified_value

from .model import *


@runtime_checkable
class IProduce(Protocol):
    """
    生产设备
    """

    log: logging.Logger

    props: Optional[ProduceProperties] = None

    def can_produce(self, *args, **kwargs) -> bool:
        """
        是否可以生产指定的产品
        """
        raise NotImplementedError()

    def on_material_ready(
        self, materials: List[Material], recipe: Optional[Recipe] = None
    ):
        """
        源材料到达事件回调
        """
        pass

    def on_product_ready(
        self, products: List[Product], recipe: Optional[Recipe] = None
    ):
        """
        产品生产完成事件回调
        """
        raise NotImplementedError()

    def perform_producing(self, *args, **kwargs) -> Generator[Any, Any, List[Product]]:
        """
        真正的生产过程，例如仿真中的timeout过程
        """
        result: List[Product] = []
        raise NotImplementedError()
        return result

    def produce(self, *args, props: Optional[ProduceProperties] = None, **kwargs):
        """
        生产入口
        """
        log = self.log
        p = props or self.props
        materials, recipe = p.materials, p.recipe

        id_recipe_map = recipe.material_dict()
        __id_materials_map = defaultdict(list)
        id_material_map: dict[str, Material] = defaultdict()

        # 把原料合并起来
        for material in materials:
            k = material.unique_id
            __id_materials_map[k].append(material)
        for k, group in __id_materials_map.items():
            id_material_map[k] = merge_quantified_value(group)

        keys = id_recipe_map.keys()

        # 计算现有原料最多能生产多少套产品
        p.product_quantity = count = min(
            [floor(id_material_map[k] / id_recipe_map[k]) for k in keys]
        )

        # 计算剩余原料
        for k, material in id_material_map.items():
            material.decrease(id_recipe_map[k].quantity * count)
        p.materials = list(id_material_map.values())

        # 发出生产完成事件
        for _ in range(count):
            products = yield from self.perform_producing(**kwargs)
            if products:
                r = self.on_product_ready(
                    products=products, recipe=recipe, order=kwargs.get("order")
                )
                if isinstance(r, types.GeneratorType):
                    yield from r
            else:
                log.error("%s produce failed", self)



from abc import ABCMeta, abstractmethod
from inspect import isclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    NewType,
    Optional,
    Type,
    Union,
)

from logis.data_type import Point

Path = List[Point]


PathFindingAlgorithmType = NewType("PathFindingAlgorithmType", str)

AlgorithmTypeResolver = Callable[[Any], PathFindingAlgorithmType]

# TODO: 完善名字、补充类型
A_STAR_ALGORITHM = PathFindingAlgorithmType("a_star")
DIJKSTRA_ALGORITHM = PathFindingAlgorithmType("dijkstra")

if TYPE_CHECKING:
    from .model import (
        PathFindingInput,
        PathFindingOutput,
    )


class PathFindingAlgorithm(metaclass=ABCMeta):
    """
    寻路算法
    """

    type: Union[PathFindingAlgorithmType, str]

    @abstractmethod
    def find_path(self, input: "PathFindingInput", **kwargs) -> "PathFindingOutput":
        """
        寻找从start到end的路径
        """
        pass


def default_algorithm_matcher(alg_name: str, **space) -> PathFindingAlgorithmType:
    """
    默认算法匹配器
    根据输入的类型匹配对应的寻路算法
    如果找不到，则返回默认的A*算法
    """
    if alg_name == "default":
        return A_STAR_ALGORITHM

    for _, v in (space or globals()).items():
        if isinstance(v, str) and v == alg_name:
            return v
    return A_STAR_ALGORITHM


def get_algorithm(
    type: PathFindingAlgorithmType, **space
) -> Type[PathFindingAlgorithm]:
    """
    根据类型获取对应的寻路算法类
    """
    # TODO：支持热加载模块
    for _, value in (space or globals()).items():
        if (
            isclass(value)
            and issubclass(value, PathFindingAlgorithm)
            and getattr(value, "type", None) == type
        ):
            return value
    raise ValueError(f"Unknown pathfinding algorithm type: {type}")


def find_algorithm(
    resolver: AlgorithmTypeResolver, space: Optional[Dict] = None, **kwargs
) -> Type[PathFindingAlgorithm]:
    """
    根据解析器和参数获取对应的寻路算法类

    Args:
        resolver: 解析器函数
        kwargs: 解析器需要的参数
    """
    type = resolver(**kwargs)
    return get_algorithm(type, **space)


def track_path(
    tree: Dict[Point, List[Point]],
    source: Point,
    target: Optional[Point] = None,
    limit=None,
) -> List[Path]:
    """
    正向生成路线，找到到终点的所有路线
    """

    paths = []

    def dfs(path: Path, root: Point):
        if (limit is not None and len(path) == limit) or (root is None):
            paths.append(path)
            return
        path.append(root)
        if root == target:
            paths.append(path)
            return
        children = tree.get(root, [])
        if not children:
            dfs(path.copy(), None)
            return
        for child in children:
            dfs(path.copy(), child)

    dfs([], source)
    return paths



import pytest
import simpy

from logis.simpy import (
    DeadlineEvent,
    DeadlineException,
    interrupt_if_timeout,
    resize_container,
    run_until,
    stop_simulation,
)


def sleep(env: simpy.Environment, n: int):
    print(f"sleep {n} start at {env.now}")
    yield env.timeout(n)
    print(f"sleep {n} end at {env.now}")


def test_until():
    env = simpy.Environment()
    env.process(sleep(env, 5))
    env.run(until=10)
    assert env.now == 10


def test_run_until():
    env = simpy.Environment()

    env.process(sleep(env, 10))
    run_until(env, max_sim_time=15)
    assert env.now == 10

    env.process(sleep(env, 5))
    run_until(env, max_sim_time=15)
    assert env.now == 15

    env.process(sleep(env, 5))
    run_until(env, max_sim_time=18)
    assert env.now == 15


def test_deadline_event():

    env = simpy.Environment()

    DeadlineEvent(env).schedule_at(15)
    env.process(sleep(env, 10))
    env.run()
    assert env.now == 15

    DeadlineEvent(env).schedule_at(20)
    env.process(sleep(env, 35))
    env.run()
    assert env.now == 20


def test_resize_container():
    env = simpy.Environment()
    container = simpy.Container(env, init=0, capacity=10)

    def get(n):
        print(f"get {n} start at {env.now}")
        yield container.get(n)
        print(f"get {n} end at {env.now}")

    def put(n):
        print(f"put {n} start at {env.now}")
        yield container.put(n)
        print(f"put {n} end at {env.now}")

    def resize_after(timeout: int, new_capacity: int):
        yield env.timeout(timeout)
        print(f"resize to {new_capacity} at {env.now}")
        resize_container(container, new_capacity)
        print(f"container capacity: {container.capacity}")

    def wait_request_done():
        while container.get_queue or container.put_queue:
            yield env.timeout(1)
            print(f"request not done, {env.now}")

    env.process(put(15))
    env.process(resize_after(5, 20))
    # env.process(wait_request_done())
    env.process(get(10))
    env.run()
    assert container.level == 5


def test_stop_simulation():
    env = simpy.Environment()
    exit_event = env.event()

    def sleep(n):
        print(f"sleep {n} start at {env.now}")
        yield env.timeout(n)
        print(f"sleep {n} end at {env.now}")

    def trigger_exit_after(timeout: int):
        yield env.timeout(timeout)
        exit_event.succeed()

    env.process(sleep(10))
    env.process(trigger_exit_after(6))
    run_until(env, max_sim_time=5, exit_signal=exit_event)
    assert env.now == 0

    run_until(env, max_sim_time=10)
    assert env.now == 6

    assert len(env._queue) > 0
    run_until(env, max_sim_time=10)
    assert env.now == 10

    env.process(sleep(5))
    stop_simulation(env)
    env.run()
    assert env.now == 10



import asyncio
from collections import defaultdict
from contextvars import ContextVar
from itertools import count
from numbers import Number
from typing import Any, Callable, List, Optional


class Context:
    """
    上下文,按照协程隔离
    """

    # _thread_local = threading.local()
    _context_vars_ = ContextVar("__logis_context_vars__", default=None)

    @classmethod
    def errors(cls) -> List[Any]:
        return cls.get("errors", default_factory=list, create=True)

    @classmethod
    def _is_in_asyncio(cls):
        try:
            task = asyncio.current_task(asyncio.get_running_loop())
            return task is not None
        except Exception as e:
            return False

    @classmethod
    def init(cls, default={}):
        if cls._context_vars_.get() is None:
            cls._context_vars_.set(default)

    @classmethod
    def set(cls, key, value):
        cls.init()
        cls._context_vars_.get()[key] = value

        # setattr(cls._thread_local, key, value)

    @classmethod
    def get(
        cls,
        key,
        default: Optional[Any] = None,
        default_factory: Callable[[], Any] = None,
        create=False,
    ):
        """
        Args:
            key: 键
            default: 默认值,已废弃,请使用default_factory
            default_factory: 默认值工厂
            create: 是否在值不存在或为空时设置默认值
        """
        dc = cls.get_all()
        result = None if dc is None else dc.get(key, None)
        if result is None:
            result = default_factory() if default_factory is not None else default
        if create:
            cls.set(key, result)
        # return getattr(cls._thread_local, key, None)
        return result

    @classmethod
    def get_all(cls) -> Optional[dict]:
        return cls._context_vars_.get()

    @classmethod
    def reset(cls):
        """
        清除上下文
        """
        if dc := cls.get_all():
            dc.clear()
            del dc
        cls._context_vars_.set(None)
        # cls._thread_local.__dict__.clear()

    @classmethod
    def count(cls, name: str, value: Number, override=False) -> Number:
        """
        计数
        Args:
            name: 待统计量
            value: 值
            override: 是否直接覆盖旧值
        Returns:
            Number: 最新值
        """
        k = "__logis_count__"
        if k not in cls.get_all():
            cls.set(k, defaultdict(float))
        if override:
            cls.get(k)[name] = value
        else:
            cls.get(k)[name] += value

        return cls.get(k)[name]

    @classmethod
    def start_timing_if_not(cls, key: str, now: float) -> bool:
        """
        如果还没有开始计时就从现在开始计时

        Args:
            key: 时长键
            now: 当前时间

        Returns:
            bool: 是否是首次开始计时
        """
        dx = cls.duration(key, now, update_end_time=False)
        return dx is None

    @classmethod
    def stop_timing(cls, key: str, now: float):
        dx = cls.duration(key, now, update_end_time=True)
        assert dx is not None, f"请先开始计时:{key}"

        return dx

    @classmethod
    def duration(
        cls, key: str, now: float = 0, update_end_time=False
    ) -> Optional[float]:
        """
        计算时长
        Args:
            key: 时长键
            now: 当前时间
            update_end_time: 是否更新结束时间

        Returns:
            float | None: 时长,如果是首次开始计时则返回None
        """
        k = "__logis_duration__"
        obj = cls.get(k, default_factory=dict, create=True)
        first_time = False
        if key not in obj:
            obj[key] = [now, None]
            first_time = True
        if update_end_time:
            obj[key][1] = now
        return None if first_time else now - obj[key][0]

    @classmethod
    def counter(cls, id: str = "_default_", start: int = 0, step: int = 1):
        """
        计数器

        Args:
            id: 计数器ID
            start: 初始值
            step: 步长
        Returns:
            count: 计数器的迭代器
        """
        k = "__logis_counter__"
        dc = cls.get(k, default_factory=dict, create=True)
        counter = dc.get(id)
        if not counter:
            counter = count(start=start, step=step)
            dc[id] = counter
        return counter



import sys
from decimal import Decimal
from fractions import Fraction
from numbers import Number
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Literal,
    NewType,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import humps
from pydantic import BaseModel, ConfigDict, Field

StringNumber = str

BaseNumberType = Union[int, float]
NumberType = Union[Decimal, int, float, Fraction, Number]

ComponentIntId = int

ComponentStrId = str

ComponentId = Union[ComponentIntId, ComponentStrId]

if sys.version_info >= (3, 9):
    from typing_extensions import TypeAlias

    TmpId: TypeAlias = Union[int, str]
else:
    TmpId = Union[int, str]


T = TypeVar("T")
Predicate = Callable[[T], bool]

Unit = NewType("Unit", str)

# 倍率计算器，输入源单位、目标单位，输出倍率
RatioComputer = Callable[[Unit, Unit], Decimal]

EventType = NewType("EventType", str)
TaskType = NewType("TaskType", str)

TuplePoint = Tuple[Optional[Number], Optional[Number], Optional[Number]]

# 这里不要随便改，如果不满足自己的需求，可以新建配置
DEFAULT_PYDANTIC_MODEL_CONFIG = ConfigDict(
    strict=False,
    arbitrary_types_allowed=True,
    extra="ignore",
    validate_by_alias=True,
    validate_by_name=True,
    coerce_numbers_to_str=True,
    alias_generator=humps.camelize,
    populate_by_name=True,
)
class CallableInput(BaseModel):
    args: Tuple = Field(default_factory=tuple)
    kwargs: Dict[str, Any] = Field(default_factory=dict)

    @staticmethod
    def of(*args, **kwargs):
        return CallableInput(args=args, kwargs=kwargs)


class InvokeResult:
    def __init__(self):
        self.is_generator = False
        self.return_value: Any = None
        self.yield_values: Any = []

    @staticmethod
    def returns(value: Any):
        self = InvokeResult()
        self.return_value = value
        return self


class Data(BaseModel, Generic[T]):
    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG
    type: str
    value: T
    tags: Optional[List[str]] = None
    description: Optional[str] = None


from abc import ABCMeta
from enum import EnumMeta


class ABCEnumMeta(EnumMeta, ABCMeta):
    """
    适用于enum类继承但是提示metaclass冲突的情况
    1. 置顶metaclass为本类即可
    """

    pass



import sys
from abc import ABCMeta
from pathlib import Path
from typing import Optional, Union

import pytest
from pydantic import AliasChoices, BaseModel, ConfigDict, Field, ValidationError

MODEL_CONFIG = ConfigDict(
    arbitrary_types_allowed=True, validate_by_alias=True, validate_by_name=True
)


class NumberUnit(metaclass=ABCMeta):
    unit: Optional[str] = None
    quantity: Union[int, float] = 1

    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class QuantifiedValue(NumberUnit, BaseModel):

    def __init__(cls, *args, **kwargs):
        super().__init__(*args, **kwargs)

    model_config = MODEL_CONFIG
    name: Optional[str] = None


class OverideQuantifiedValue(QuantifiedValue):

    unit: str = Field(validation_alias=AliasChoices("xxx", "unit"))

    model_config = MODEL_CONFIG


class Xxx(OverideQuantifiedValue):
    pass


Yyy = OverideQuantifiedValue


def test_override_quantified_value():
    x = OverideQuantifiedValue.model_validate(dict(quantity=10, xxx="m"))
    assert x.quantity == 10
    assert x.unit == "m"

    with pytest.raises(AssertionError):

        x = Xxx.model_validate(dict(quantity=10, xxx="m"))
        assert x.quantity == 10
        assert x.unit == "m"

    y = Yyy.model_validate(dict(quantity=10, xxx="m"))
    assert y.quantity == 10
    assert y.unit == "m"


class ComplexClass:

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.name: str = "hello"


def test_mro():
    class Combined2(QuantifiedValue, ComplexClass):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)

    class Combined(ComplexClass, QuantifiedValue):
        def __init__(self, **kwargs):
            super().__init__(**kwargs)
            # QuantifiedValue.__init__(self, **kwargs)
            # ComplexClass.__init__(self, **kwargs)

    from logis.util.visual_util import draw_mro_tree

    draw_mro_tree(Combined, Combined2)

    c = Combined()
    assert c.name == "hello"

    with pytest.raises(AssertionError):
        c2 = Combined2()
        assert c2.name == "hello"


def test_path():
    class MyModel(BaseModel):
        path: Optional[Path] = None

    m = MyModel.model_validate(dict(path="."))
    assert m.path.exists()

class TestModel(BaseModel):
    name: Optional[str] = Field(None, validation_alias=AliasChoices("名称"))
    age: Optional[int] = Field(None, validation_alias=AliasChoices("年龄", "age"))

    model_config = ConfigDict(extra="ignore")


def test_validation_alias_choices():
    form = dict(名称="张三", age="18", not_exist=True)
    m = TestModel.model_validate(form)
    assert m.name == "张三"
    assert m.age == 18

    form = dict(name="张三", 年龄="18")
    if sys.version_info >= (3, 9):
        m = TestModel.model_validate(form, by_alias=True, by_name=True)
        assert m.name == "张三"
    else:
        m = TestModel.model_validate(form)
        assert m.name is None

    assert m.age == 18

    # with pytest.raises(ValidationError):
    #     form = dict(name="张三", 年龄="18")
    #     m = TestModel.model_validate(form)
    #     if sys.version_info>=(3,9):
    #         assert m.name == "张三"
    #     else:
    #         assert m.name is None
    #     assert m.age == 18


def test_implict_type_cast():
    class MyModel(BaseModel):
        name: str
        model_config = ConfigDict(strict=False, coerce_numbers_to_str=True)

        age: Union[int, float, None] = None

    v = MyModel(name=1000, age=100.0)
    assert v.name == "1000"
    assert type(v.age) is float and v.age == 100



from typing import Protocol, runtime_checkable

from logis.data_type import Point
from logis.data_type.unitable import TIME_UNIT_CONFIG, UnitConfigBuilder


def test_point():
    p1 = Point(1, 2)
    p2 = Point(1, 2, 3)
    p3 = Point([1, 2, 3])
    p4 = Point([10, 1])
    for p in [p1, p2, p3, p4]:
        print(p)

    assert p1 - p2 == Point(0, 0, -3)
    assert p3 - p4 == Point(-9, 1, 3)
    assert Point() - Point() == Point()


def test_unit_config_builder():
    unit_config = UnitConfigBuilder.with_base("m").add("cm", 100).build()
    ratio = unit_config.get_ratio("m", "cm")
    assert ratio == 100
    ratio = unit_config.get_float_ratio("cm", "m")
    assert ratio == 0.01
    ratio = unit_config.get_ratio("m", "m")
    assert ratio == 1

    x = TIME_UNIT_CONFIG | unit_config

    ratio = x.get_ratio("秒", "毫秒")
    assert ratio == 1000


def test_protocal():
    @runtime_checkable
    class MyProtocal(Protocol):
        pass

    class SomeClass:
        pass

    assert isinstance(SomeClass(), MyProtocal)



from abc import ABCMeta, abstractmethod


class ILifecycle(metaclass=ABCMeta):
    """
    生命周期接口
    """

    @abstractmethod
    def on_start(self, **kwargs):
        """
        启动
        """

    @abstractmethod
    def on_stop(self, **kwargs):
        """
        停止
        """

    @abstractmethod
    def on_restart(self, **kwargs):
        """
        重启
        """



import pandas as pd


def test_cascade():
    # 外层数据
    outer_df = pd.DataFrame({"id": [1, 2], "name": ["A", "B"]})

    # 内层数据（通过 id 与外层关联）
    inner_df = pd.DataFrame(
        {
            "id": [1, 1, 2, 2],  # 关联键
            "param": ["x", "y", "x", "y"],
            "value": [10, 20, 30, 40],
        }
    )

    # 通过 merge 关联内外层数据
    a = pd.merge(outer_df, inner_df, on="id")
    print(a)

    b = pd.json_normalize(
        data={
            "id": [1, 2],
            "name": ["A", "B"],
            "params": [
                {"key": "x", "value": 10},
                {"key": "y", "value": 20},
            ],
            "metadata": {
                "created_at": "2023-01-01",
                "updated_at": "2023-01-02",
            },
        }
    )
    print(b)



__doc__ = """
配置模块
包含一些全局配置
"""


import logging
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional


class IAppConfig(ABC):
    """
    应用配置接口
    """

    @abstractmethod
    def get_runtime_root_dir(self, *args, **kwargs) -> Optional[Path]:
        pass

    def get_root_dir(self, *args, **kwargs) -> Optional[Path]:
        """
        此方法只是一个别名，返回与get_runtime_root_dir相同的结果
        """
        return self.get_runtime_root_dir(*args, **kwargs)

    # @abstractmethod
    # def get_user_data_root_dir(self, *args, **kwargs) -> Optional[Path]:
    #     pass

    @abstractmethod
    def get_logs_dir(self, *args, **kwargs) -> Optional[Path]:
        pass

    @abstractmethod
    def get_config_dir(self, *args, **kwargs) -> Optional[Path]:
        pass

    @abstractmethod
    def get_data_dir(self, *args, **kwargs) -> Optional[Path]:
        pass

    def get_temp_dir(self, *args, **kwargs) -> Optional[Path]:
        raise NotImplementedError()

    def get_middleware_dir(self, *args, **kwargs) -> Optional[Path]:
        raise NotImplementedError()

    def get_extension_dir(self, *args, **kwargs) -> Optional[Path]:
        raise NotImplementedError()

    def get_app_logger(self, *args, **kwargs) -> logging.Logger:
        raise NotImplementedError()

    def get_sidecar_dir(self, *args, **kwargs) -> Optional[Path]:
        raise NotImplementedError()



import enum
from typing import List, Literal, Optional, TypeAlias

from pydantic import BaseModel, Field


class ComponentConfigItem(BaseModel):
    """
    组件配置项，用于前端展示和后端处理
    """

    # 中文名
    label: Optional[str] = None
    # 描述
    description: Optional[str] = None
    # 类型编码
    type_id: str
    # 加载器，用于指导后端解析组件
    loader: Optional[str] = None
    # 是否卸载，用于卸载已加载的组件
    unload: bool = False

    # 是否启用配置
    enabled: bool = True


class ComponentConfig(BaseModel):
    """
    组件配置
    """

    entity_list: List[ComponentConfigItem] = Field(default_factory=list)


class BlueprintKind(enum.Enum):
    """
    蓝图的类型，如果不够用，可以修改此类或者继承并添加新的类型
    """

    SOURCE = "source"
    ENTER = "enter"
    EXIT = "exit"
    SINK = "sink"
    ASSEMBLER = "assembler"
    DELAY = "delay"
    EVENT = "event"
    GOODS_TO_PERSON = "goods_to_person"
    PRODUCER = "producer"
    WORK_STATION = "work_station"

    RESOURCE_POOL = "resource_pool"
    TRANSPORT = "transport"

    RACK_SYSTEM = "rack_system"
    RACK = "rack"
    AGV_RACK = "agv_rack"
    PALLET_RACK = "pallet_rack"
    PICKING_RACK = "picking_rack"
    ASRS = "asrs"
    REPOSITORY = "repository"

    CONVEYOR_BELT = "conveyor_belt"
    TURNING_STATION = "turning_station"
    PATH = "path"
    GRID = "grid"
    POINT_NODE = "point_node"
    RECTANGULAR_NODE = "rectangular_node"

    AGENT = "agent"
    AGV = "agv"
    STOCK = "stock"
    FORKLIFT = "forklift"



import pytest


def test_compare_counter_progress():
    from collections import Counter

    from logis.util.dict_util import compare_counter_progress

    progress = Counter(a=1, b=2, c=3)
    target = Counter(a=1, b=2, c=3, d=4)
    assert compare_counter_progress(progress, target) < 0
    progress["c"] = 100

    assert compare_counter_progress(progress, target) < 0

    progress["d"] = 100
    assert compare_counter_progress(progress, target) > 0



import threading
import time
from functools import reduce
from threading import Thread


def test_cost():
    def run(dt: int):
        return reduce(lambda x, y: x * y, range(1, 10000))

    start = time.time()
    run(0)
    dt = time.time() - start
    print(dt)

    ts = []
    for i in range(20):
        t = Thread(target=run, args=(i,), name=str(i))
        ts.append(t)
    start = time.time()
    for t in ts:
        t.start()
        t.join()
    dt = time.time() - start
    print(dt)



from typing import Any, Callable

ErrorHandler = Callable[[Exception], Any]


class MyBaseException(Exception):
    """
    基础异常类，所有自定义异常都应该继承自这个类
    """

    def __init__(self, *args, **attrs):
        super().__init__(*args)
        for k, v in attrs.items():
            hasattr(self, k) or setattr(self, k, v)


class UserFacingException(MyBaseException):
    """
    用户面向的异常类，用于在前端展示异常信息
    """

    def __init__(self, *args, **attrs):
        super().__init__(*args, **attrs)



from typing import Any, Tuple

from pydantic import BaseModel, Field

from .base import DEFAULT_PYDANTIC_MODEL_CONFIG


class EchartsConfig(BaseModel):
    option: dict = Field(default_factory=dict)
    attrs: dict = Field(default_factory=dict)

    model_config = DEFAULT_PYDANTIC_MODEL_CONFIG

    def add_class(self, *class_list: str):
        self.attrs.setdefault("class", []).extend(class_list)
        return self

    def add_css_style(self, *pairs: Tuple[str, Any]):
        self.attrs.setdefault("style", {}).update({k: v for k, v in pairs})
        return self



import asyncio
from typing import Union

import fastapi
import websockets
from websockets.asyncio.client import connect

try:
    from websockets.client import ClientConnection
    from websockets.server import ServerConnection
except:
    from websockets import ClientConnection, ServerConnection


WebSocketConnection = Union[ServerConnection, fastapi.WebSocket, ClientConnection]


def is_websocket_closed(conn: WebSocketConnection):
    """
    判断WebSocket连接是否已关闭
    """
    if isinstance(conn, fastapi.WebSocket):
        from starlette.websockets import WebSocketState

        return conn.client_state == WebSocketState.DISCONNECTED
    if isinstance(conn, (ClientConnection)):
        return conn.state == websockets.State.CLOSED
    if isinstance(conn, ServerConnection):
        return conn.state == websockets.State.CLOSED

    raise ValueError(f"Unsupported WebSocket connection type: {type(conn)}")


async def send_text(conn: WebSocketConnection, data: str):
    if isinstance(conn, ServerConnection):
        await conn.send(data)
    elif isinstance(conn, fastapi.WebSocket):
        await conn.send_text(data)
    else:
        raise NotImplementedError(
            f"send_text not implemented for this connection type: {type(conn)}"
        )


def get_close_code(conn: WebSocketConnection):
    if isinstance(conn, ServerConnection):
        return conn.close_code
    elif isinstance(conn, fastapi.WebSocket):
        return None


def parse_client_info(connection: WebSocketConnection):
    """
    从连接中解析客户端ip、port等信息
    """
    if isinstance(connection, ServerConnection):
        client_ip, client_port = connection.transport.get_extra_info("peername")
    elif isinstance(connection, fastapi.WebSocket):
        client_ip, client_port = connection.client.host, connection.client.port
    else:
        client_ip, client_port = None, None
    return client_ip, client_port


def get_write_buffer_size(connection: WebSocketConnection):
    """
    获取连接的写缓冲区大小
    """
    if isinstance(connection, ServerConnection):
        return connection.transport.get_write_buffer_size()
    else:
        raise NotImplementedError(
            "get_write_buffer_size not implemented for this connection type"
        )


def get_write_buffer_limits(connection: WebSocketConnection):
    """
    获取连接的写缓冲区可用大小
    """
    if isinstance(connection, ServerConnection):
        return connection.transport.get_write_buffer_limits()
    else:
        raise NotImplementedError(
            "get_write_buffer_limits not implemented for this connection type"
        )


async def interactive_client():
    url = input("input websocket server url: ").strip()
    url = url if url.startswith("ws://") or url.startswith("wss://") else f"ws://{url}"
    async with connect(url) as conn:
        # 创建一个任务来接收消息
        async def receive_messages():
            while True:
                try:
                    msg = await conn.recv()
                    print(f"\nReceived: {msg}")
                    print("Enter message (or 'quit' to exit): ", end="", flush=True)
                except websockets.exceptions.ConnectionClosed:
                    print("\nConnection closed by server")
                    return

        # 创建一个任务来发送消息
        async def send_messages():
            while True:
                try:
                    message = await asyncio.get_event_loop().run_in_executor(
                        None, input, "Enter message (or 'quit' to exit): "
                    )
                    if message.lower() == "quit":
                        print("Closing connection...")
                        await conn.close()
                        return
                    await conn.send(message)
                except websockets.exceptions.ConnectionClosed:
                    print("\nConnection closed")
                    return

        # 同时运行接收和发送任务
        receive_task = asyncio.create_task(receive_messages())
        send_task = asyncio.create_task(send_messages())

        # 等待任一任务完成
        done, pending = await asyncio.wait(
            [receive_task, send_task], return_when=asyncio.FIRST_COMPLETED
        )

        # 取消未完成的任务
        for task in pending:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                pass


def start_interactive_client():
    asyncio.run(interactive_client())


if __name__ == "__main__":
    start_interactive_client()



from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler

from .builder import *
from .fmt import *
from .handler import *
from .util import *


def get_default_dict_config_builder():
    """
    默认内置
    1. 名为console的控制台Logger，日志级别INFO，彩色输出
    2. 名为app的Logger，日志级别INFO，输出到控制台和文件
    """
    return (
        DictConfigBuilder()
        .handler(get_default_console_handler(), name="console")
        .handler_dict(
            HandlerDictConfig(
                name="error_file_handler",
                clazz=RotatingFileHandler,
                level=logging.ERROR,
                filename="error.log",
                formatter=LoggerFormat.DEFAULT.formatter_name(),
            )
        )
        .handler_dict(
            HandlerDictConfig(
                name="info_file_handler",
                clazz=TimedRotatingFileHandler,
                level=logging.INFO,
                filename="info.log",
                formatter=LoggerFormat.DEFAULT.formatter_name(),
            )
        )
        .logger_dict(
            LoggerDictConfig(
                name="root",
                level=logging.WARNING,
                handlers=["console", "error_file_handler"],
            )
        )
        .logger_dict(
            LoggerDictConfig(
                name="app",
                level=logging.INFO,
                handlers=["console", "info_file_handler", "error_file_handler"],
            )
        )
    )



import asyncio
import atexit
import inspect
import logging
import queue
import threading
import time
from abc import ABC, abstractmethod
from typing import (
    Any,
    Callable,
    Coroutine,
    Generic,
    List,
    Literal,
    Optional,
    Type,
    TypeVar,
    Union,
)

from pydantic import BaseModel, ConfigDict

from logis.iface import QueueType
from logis.util.lambda_util import invoke

T = TypeVar("T")


def get_or_new_event_loop(auto_set: bool = True) -> asyncio.AbstractEventLoop:
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        if auto_set:
            asyncio.set_event_loop(loop)
    return loop


class WriteBufferConfig(BaseModel):
    # 每批最大事件数
    batch_size: int = 256

    # 缓存的最大条目数，默认无限制
    queue_size: int = 0

    # 缓存的最大时间间隔
    flush_interval: float = 10

    queue_type: Optional[QueueType] = None

    _queue: Optional[QueueType] = None

    handler: Union[Callable[[List[T]], None], None, Coroutine[Any, Any, None]] = None

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def get_queue(self, mode: Literal["sync", "async"]) -> QueueType:
        if self._queue is None:
            if self.queue_type is None:
                self.queue_type = queue.Queue if mode == "sync" else asyncio.Queue
            self._queue = self.queue_type(maxsize=self.queue_size)
        return self._queue


class WriteBuffer(ABC, Generic[T]):
    """
    写操作缓冲器
    """

    def __init__(self, config: WriteBufferConfig, **kwargs):
        self._config = config
        self._handler = self._config.handler

        self._is_running = False
        self._last_flush_time = time.time()
        self._lock: Union[threading.Lock, asyncio.Lock, None] = None
        self._flush_task: Optional[threading.Thread] = None

    def need_flush(self) -> bool:
        """判断是否需要刷新（根据时间间隔和条目数）"""
        current_time = time.time()
        meet_time_interval = (
            current_time - self._last_flush_time >= self._config.flush_interval
        )
        if meet_time_interval:
            logging.debug(f"meet time interval {self._config.flush_interval}")
        meet_batch_size = self._queue.qsize() >= self._config.batch_size
        if meet_batch_size:
            logging.debug(f"meet batch size {self._config.batch_size}")
        return meet_time_interval or meet_batch_size

    def put(self, item: Any) -> None:
        """添加单条数据到缓冲器"""
        with self._lock:
            # 若队列满了，先触发一次刷新（避免阻塞）
            if self._queue.full():
                self.flush()
            self._queue.put(item)

        # 检查是否达到批量阈值，达到则触发刷新
        if self.need_flush():
            self.flush()

    def start(self) -> None:
        """启动缓冲器（开启定时刷新线程）"""
        self._is_running = True
        self._last_flush_time = time.time()

    @abstractmethod
    def stop(self) -> None:
        """停止缓冲器（强制刷新剩余数据并关闭线程）"""
        self._is_running = False

    @abstractmethod
    def flush(self, flush_all: bool = False, **kwargs):
        pass

    def _flush_loop(self, **kwargs) -> None:
        """定时刷新循环（后台线程）"""
        pass


class SyncWriteBuffer(WriteBuffer[T]):
    """
    同步写操作缓冲器
    """

    def __init__(self, config: WriteBufferConfig, **kwargs):
        super().__init__(config, **kwargs)
        self._queue = self._config.get_queue(mode="sync")
        self._lock = threading.Lock()

        atexit.register(self.stop)

    def start(self):
        if self._is_running:
            return
        super().start()
        t = threading.Thread(target=self._flush_loop, daemon=True)
        t.start()
        return t

    def stop(self) -> None:
        """停止缓冲器（强制刷新剩余数据并关闭线程）"""
        super().stop()
        if self._flush_task:
            self._flush_task.join()
        self.flush(flush_all=True)  # 退出前强制刷新剩余数据

    def flush(self, flush_all: bool = False, **kwargs) -> None:
        """强制刷新缓冲数据（批量处理）"""
        if self._queue.empty():
            return

        # TODO：是否可以不用加锁
        items: List[Any] = []
        with self._lock:
            while not self._queue.empty() and (
                flush_all or len(items) < self._config.batch_size
            ):
                try:
                    items.append(self._queue.get_nowait())
                except queue.Empty:
                    break

        # 调用用户自定义的批量处理函数
        if items:
            self._handler(items)
            # 更新刷新时间
            self._last_flush_time = time.time()

    def _flush_loop(self, check_interval=0.1, **kwargs):
        while self._is_running:
            if self.need_flush():
                self.flush()
            time.sleep(check_interval)


class AsyncWriteBuffer(WriteBuffer[T]):
    """
    异步写操作缓冲器
    """

    def __init__(self, config, **kwargs):
        super().__init__(config, **kwargs)
        self._loop = get_or_new_event_loop()
        self._queue = self._config.get_queue(mode="async")
        self._lock = asyncio.Lock()
        atexit.register(self.__wait_stop)

    def __wait_stop(self):
        if self._loop.is_closed():
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        else:
            loop = self._loop

        loop.run_until_complete(self.stop())

    async def start(self):
        if self._is_running:
            return
        super().start()
        self._flush_task = self._loop.create_task(self._flush_loop())
        return self._flush_task

    async def stop(self):
        super().stop()
        if self._flush_task:
            await self._flush_task
        await self.flush(flush_all=True)

    async def put(self, item: Any):
        async with self._lock:
            if self._queue.full():
                await self.flush()
            await self._queue.put(item)

        if self.need_flush():
            await self.flush()

    async def flush(self, flush_all=False, **kwargs):
        if self._queue.empty():
            return

        items: List[Any] = []
        # TODO：是否可以不用加锁
        async with self._lock:
            while not self._queue.empty() and (
                flush_all or len(items) < self._config.batch_size
            ):
                items.append(await self._queue.get())

        if items:
            if inspect.iscoroutinefunction(self._handler):
                await self._handler(items)  # 调用异步批量处理函数
            else:
                self._handler(items)
            self._last_flush_time = time.time()

    async def _flush_loop(self, check_interval=0.1, **kwargs):
        while self._is_running:
            if self.need_flush():
                await self.flush()
            await asyncio.sleep(check_interval)
            # TODO: 这里会出现乱序问题
            # try:
            #     item = await asyncio.wait_for(self._queue.get(), timeout=check_interval)
            #     self._queue.put_nowait(item)
            # except asyncio.TimeoutError:
            #     pass



import inspect
from itertools import chain
from math import e
from typing import Iterable

import matplotlib.pyplot as plt
import networkx as nx
from networkx.drawing.nx_agraph import graphviz_layout


def mro_tree(cls):
    """把 inspect.getclasstree 结果扁平化并返回父->子边"""
    tree = inspect.getclasstree(inspect.getmro(cls))
    edges = []

    def walk(node):
        if not node:
            return
        if not isinstance(node, Iterable):
            return
        if isinstance(node, list):  # 同级列表
            for item in node:
                walk(item)
        elif len(node) == 2:
            # node = (class, (parent_list))
            cls, bases = node
            if not isinstance(bases, Iterable):
                edges.append((cls, None))  # 无父类的类
                return
            for p in bases:
                edges.append((p, cls))  # 父 -> 子
            walk(list(bases))
        else:
            pass

    walk(tree)
    return edges


def draw_mro_tree(*cls_list, use_graphviz=False):
    for cls in cls_list:
        edges = mro_tree(cls)
        G = nx.DiGraph()
        G.add_edges_from(edges)

        plt.figure(figsize=(6, 4))
        if use_graphviz:

            pos = graphviz_layout(G, prog="dot")
        else:
            pos = nx.spring_layout(G, seed=42)

        nx.draw(
            G,
            pos,
            with_labels=True,
            node_size=2000,
            node_color="skyblue",
            font_size=10,
            arrowsize=15,
        )
        plt.axis("off")
        plt.show()



from typing import Union

import simpy

from .exception import DeadlineException
from .util import schedule_event_at


class DeadlineEvent(simpy.Event):
    """
    截止事件，用于在指定时间触发**DeadlineException**异常从而正常退出
    如果仿真早于截止时间完成, 则此事件仍会执行
    """

    def interrupt(self, *args, **kwargs):
        """
        中断事件
        """
        # 如果后续还有事件,忽略事件正常退出,此时时钟停止在deadline时间
        raise DeadlineException(f"deadline {self.at} reached, force exit")

    def __init__(
        self,
        env: simpy.Environment,
        at: Union[float, int, None] = None,
    ):
        super().__init__(env)
        self.at: Union[float, int, None] = None
        self._ok = True
        self._value = None
        if at is not None:
            self.schedule_at(at)
        # self.callbacks = [StopSimulation.callback]
        self.callbacks = [self.interrupt]

    def schedule_at(self, at: Union[float, int]):
        """
        计划在指定时间触发事件
        Args:
            at (float | int): 触发时间
        """
        if self.at is not None:
            raise ValueError("deadline event already scheduled")
        self.at = at
        return schedule_event_at(self.env, at, self)



import logging
import sys
from collections import defaultdict
from logging import StreamHandler
from logging.handlers import RotatingFileHandler, TimedRotatingFileHandler
from pathlib import Path
from typing import Dict, List, Optional, Type, Union
from uuid import uuid4

from platformdirs import user_data_dir

from .util import add_formatter_if_not, format_filename, format_level


class LoggerBuilder:
    """
    日志构建器接口
    """

    def __init__(self, **kwargs) -> None:
        self.dir(Path(user_data_dir()) / "logs").name("root").level(
            logging.WARNING
        ).encoding("utf-8").backup_count(5).rotation_when("h").max_bytes(
            200 * 1024 * 1024
        ).propagate(
            False
        )

        self._handlers: Dict[str, logging.Handler] = defaultdict()

    def _resolve_file_path(self, filename: str) -> Optional[Path]:
        if not filename:
            return None
        return Path(format_filename(filename, self._log_dir))

    def name(self, name: str) -> "LoggerBuilder":
        """
        设置日志名称
        """
        self._name = name
        return self

    def propagate(self, propagate: bool) -> "LoggerBuilder":
        """
        设置是否 Propagate 日志到父 Logger
        """
        self._propagate = propagate
        return self

    def dir(self, path: Union[str, Path]) -> "LoggerBuilder":
        """
        设置日志目录
        """
        self._log_dir = path if isinstance(path, Path) else Path(path)
        self._log_dir.mkdir(parents=True, exist_ok=True)
        logging.debug("Log directory set to: %s", str(self._log_dir))
        return self

    def level(self, level: int = logging.WARNING) -> "LoggerBuilder":
        """
        设置root日志级别
        """
        level = format_level(level)
        self._level = level
        return self

    def max_bytes(self, max_bytes: int) -> "LoggerBuilder":
        """
        设置日志文件最大字节数
        """
        self._max_bytes = max_bytes
        return self

    def encoding(self, encoding: str) -> "LoggerBuilder":
        """
        设置日志文件编码
        """
        self._encoding = encoding
        return self

    def backup_count(self, count: int) -> "LoggerBuilder":
        """
        设置日志文件最大备份数量
        """
        self._backup_count = count
        return self

    def rotation_when(self, when: str) -> "LoggerBuilder":
        """
        设置日志文件轮转时间间隔单位
        """
        self._rotation_when = when
        return self

    def add_handler(
        self,
        handler_type: Union[Type[logging.Handler], logging.Handler],
        level: Optional[int] = None,
        filename: Optional[str] = None,
        name: Optional[str] = None,
        filters: Optional[List[logging.Filter]] = None,
        **kwargs,
    ) -> "LoggerBuilder":
        """
        添加日志处理器

        Args:
            handler_type: 日志处理器类型
            level: 日志级别
            filename: 日志文件名
            name: 日志处理器名称,如果未提供,则使用文件名或随机UUID
            filters: 日志过滤器列表
        """
        filename = self._resolve_file_path(filename)

        is_handler_instance = isinstance(handler_type, logging.Handler)
        default_handler_name = None
        if is_handler_instance:
            default_handler_name = handler_type.name
        handler_name = str(name or default_handler_name or filename or uuid4())

        if handler_name in self._handlers:
            logging.warning(
                "Handler with name %s already exists, will duplicate.", handler_name
            )

        if isinstance(handler_type, logging.FileHandler) or issubclass(
            handler_type, logging.FileHandler
        ):
            kwargs.setdefault("delay", True)
            kwargs.update(encoding=self._encoding, filename=filename)
        if handler_type is TimedRotatingFileHandler:
            handler = TimedRotatingFileHandler(
                when=self._rotation_when,
                backupCount=self._backup_count,
                **kwargs,
            )
        elif handler_type is RotatingFileHandler:
            handler = RotatingFileHandler(
                maxBytes=self._max_bytes,
                backupCount=self._backup_count,
                **kwargs,
            )
        elif handler_type is StreamHandler:
            handler = StreamHandler(stream=kwargs.get("stream", sys.stdout))
        elif is_handler_instance:
            handler = handler_type
        elif issubclass(handler_type, logging.Handler):
            handler = handler_type(**kwargs)
        else:
            raise ValueError(f"Unknown handler type: {handler_type}")

        add_formatter_if_not(handler)

        level = level if level is not None else self._level
        handler.setLevel(format_level(level))

        handler.name = handler_name

        for f in filters or []:
            handler.addFilter(f)

        self._handlers[handler_name] = handler
        return self

    def build(self) -> logging.Logger:
        assert self._name, "Logger name must be provided."
        name = self._name if self._name is not None else "root"
        logger = logging.getLogger(name)
        if self._level is not None:
            logger.setLevel(self._level)
        logger.propagate = self._propagate
        for _, handler in self._handlers.items():
            if handler is None:
                continue
            logger.addHandler(handler)
        return logger



import logging
import time
from functools import wraps


def log_execution_time(func):
    """
    打印执行事件
    TODO：考虑生成器方法
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        logging.debug(f"{func.__name__} took {end_time - start_time:.4f} seconds")
        return result

    return wrapper



from copy import copy, deepcopy


class MyInnerClass:
    def __init__(self):
        self.inner_name = "who"


class MyClass:

    def __init__(self):
        self.name: str = None
        self.age = 10
        self.inner = MyInnerClass()

    def yield_method(self):
        yield 1


def test_copy():
    obj = MyClass()
    obj1 = copy(obj)
    obj.name = 11
    assert obj1.name != obj.name
    assert obj1.inner is obj.inner


def test_deepcopy():
    obj = MyClass()
    obj2 = deepcopy(obj)
    obj.age = 1000
    assert obj.name == obj2.name
    assert obj.age != obj2.age
    assert obj.inner is not obj2.inner



from abc import ABC, abstractmethod
from typing import Any, Generator, Optional, TypeVar, Union

import simpy

from logis.biz.sim import AgentId, IBlueprint
from logis.biz.sim.transport import ITransport
from logis.data_type import Point, Speed


class IAgent(ITransport, IBlueprint):

    @property
    @abstractmethod
    def id(self) -> AgentId:
        pass

    @property
    @abstractmethod
    def rack_as_obstacle(self) -> bool:
        """
        是否将其余货架视为障碍物
        """
        pass

    @property
    @abstractmethod
    def agent_as_obstacle(self) -> bool:
        """
        是否将其余智能体视为障碍物
        """
        pass

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.origin_location: Optional[Point] = None
        self.current_location: Optional[Point] = None

    @property
    @abstractmethod
    def is_free(self, **kwargs) -> bool:
        """
        判断是否空闲
        """
        pass

    @abstractmethod
    def stop_working(self, *args, **kwargs):
        pass

    @abstractmethod
    def start_working(self, *args, **kwargs):
        pass

    @abstractmethod
    def do_move(
        self, next_point: Point, *args, **kwargs
    ) -> Generator[simpy.Event, Any, None]:
        """
        从当前位置移动到下一个相邻的位置
        """
        pass

    @abstractmethod
    def move(
        self, target: Point, speed: Optional[Union[float, Speed]] = None, **kwargs
    ) -> Generator[simpy.Event, Any, None]:
        """
        从当前位置移动到目标位置，此方法是最顶层的移动方法，内部处理路径规划并逐步调用do_move
        TODO: 考虑将speed等参数作为agent的内部属性，而不是在这里传参
        """
        pass


AgentClass = TypeVar("AgentClass", bound=IAgent)
